from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_iq.protocolentities import ResultIqProtocolEntity
class SuccessRemoveParticipantsIqProtocolEntity(ResultIqProtocolEntity):
    '''
    <iq type="result" from="{{group_jid}}" id="{{id}}">
        <remove type="success" participant="{{jid}}"></remove>
        <remove type="success" participant="{{jid}}"></remove>
    </iq>
    '''

    def __init__(self, _id, groupId, participantList):
        super(SuccessRemoveParticipantsIqProtocolEntity, self).__init__(_from = groupId, _id = _id)
        self.setProps(groupId, participantList)

    def setProps(self, groupId, participantList):
        self.groupId = groupId
        self.participantList = participantList
        self.action = 'remove'

    def getAction(self):
        return self.action

    def toProtocolTreeNode(self):
        node = super(SuccessRemoveParticipantsIqProtocolEntity, self).toProtocolTreeNode()
        participantNodes = [
            ProtocolTreeNode("remove", {
                "type":                     "success",
                "participant":       participant
            })
            for participant in self.participantList
        ]
        node.addChildren(participantNodes)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(SuccessRemoveParticipantsIqProtocolEntity, SuccessRemoveParticipantsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = SuccessRemoveParticipantsIqProtocolEntity
        participantList = []
        for participantNode in node.getAllChildren():
            if participantNode["type"]=="success":
                participantList.append(participantNode["participant"])
        entity.setProps(node.getAttributeValue("from"), participantList)
        return entity



import urllib,sys, os, logging
import hashlib
from .waresponseparser import ResponseParser
from yowsup.env import AndroidYowsupEnv
CURRENT_ENV = AndroidYowsupEnv()

if sys.version_info < (3, 0):
    import httplib
    from urllib import urlencode

    if sys.version_info >= (2, 7, 9):
        #see https://github.com/tgalal/yowsup/issues/677
        import ssl
        ssl._create_default_https_context = ssl._create_unverified_context

else:
    from http import client as httplib
    from urllib.parse import urlencode

logger = logging.getLogger(__name__)


class WARequest(object):

    OK = 200

    def __init__(self):

        self.pvars = []
        self.port = 443
        self.type = "GET"
        self.parser = None
        self.params = []
        self.headers = {}

        self.sent = False
        self.response = None



    def setParsableVariables(self, pvars):
        self.pvars = pvars

    def onResponse(self, name, value):
        if name == "status":
            self.status = value
        elif name == "result":
            self.result = value

    def addParam(self,name,value):
        self.params.append((name,value))

    def removeParam(self, name):
        for i in range(0, len(self.params)):
            if self.params[i][0] == name:
                del self.params[i]


    def addHeaderField(self, name, value):
        self.headers[name] = value

    def clearParams(self):
        self.params = []

    def getUserAgent(self):
        return CURRENT_ENV.getUserAgent()

    def send(self, parser = None):

        if self.type == "POST":
            return self.sendPostRequest(parser)

        return self.sendGetRequest(parser)

    def setParser(self, parser):
        if isinstance(parser, ResponseParser):
            self.parser = parser
        else:
            logger.error("Invalid parser")

    def getConnectionParameters(self):

        if not self.url:
            return "", "", self.port

        try:
            url = self.url.split("://", 1)
            url = url[0] if len(url) == 1 else url[1]

            host, path = url.split('/', 1)
        except ValueError:
            host = url
            path = ""

        path = "/" + path

        return host, self.port, path

    def sendGetRequest(self, parser = None):
        self.response = None
        params =  self.params#[param.items()[0] for param in self.params];

        parser = parser or self.parser or ResponseParser()

        headers = dict(list({"User-Agent":self.getUserAgent(),
                "Accept": parser.getMeta()
            }.items()) + list(self.headers.items()));

        host,port,path = self.getConnectionParameters()

        self.response = WARequest.sendRequest(host, port, path, headers, params, "GET")

        if not self.response.status == WARequest.OK:
            logger.error("Request not success, status was %s"%self.response.status)
            return {}

        data = self.response.read()
        logger.info(data)

        self.sent = True
        return parser.parse(data.decode(), self.pvars)

    def sendPostRequest(self, parser = None):
        self.response = None
        params =  self.params #[param.items()[0] for param in self.params];

        parser = parser or self.parser or ResponseParser()

        headers = dict(list({"User-Agent":self.getUserAgent(),
                "Accept": parser.getMeta(),
                "Content-Type":"application/x-www-form-urlencoded"
            }.items()) + list(self.headers.items()))

        host,port,path = self.getConnectionParameters()
        self.response = WARequest.sendRequest(host, port, path, headers, params, "POST")


        if not self.response.status == WARequest.OK:
            logger.error("Request not success, status was %s" % self.response.status)
            return {}

        data = self.response.read()

        logger.info(data)

        self.sent = True
        return parser.parse(data.decode(), self.pvars)


    @staticmethod
    def sendRequest(host, port, path, headers, params, reqType="GET"):

        params = urlencode(params)


        path = path + "?"+ params if reqType == "GET" and params else path

        if len(headers):
            logger.debug(headers)
        if len(params):
            logger.debug(params)

        logger.debug("Opening connection to %s" % host);
        conn = httplib.HTTPSConnection(host ,port) if port == 443 else httplib.HTTPConnection(host ,port)

        logger.debug("Sending %s request to %s" % (reqType, path))
        conn.request(reqType, path, params, headers);

        response = conn.getresponse()
        return response



from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
import time

class SyncIqProtocolEntity(IqProtocolEntity):

    '''
    <iq type="get" id="{{id}}" xmlns="urn:xmpp:whatsapp:sync">
        <sync
            sid="{{str((int(time.time()) + 11644477200) * 10000000)}}"
            index="{{0 | ?}}"
            last="{{true | false?}}"
        >
        </sync>
    </iq>
    '''

    def __init__(self, _type, _id = None, sid = None, index = 0, last = True):
        super(SyncIqProtocolEntity, self).__init__("urn:xmpp:whatsapp:sync", _id = _id, _type = _type)
        self.setSyncProps(sid, index, last)

    def setSyncProps(self, sid, index, last):
        self.sid = sid if sid else str((int(time.time()) + 11644477200) * 10000000)
        self.index = int(index)
        self.last = last


    def __str__(self):
        out  = super(SyncIqProtocolEntity, self).__str__()
        out += "sid: %s\n" % self.sid
        out += "index: %s\n" % self.index
        out += "last: %s\n" % self.last
        return out

    def toProtocolTreeNode(self):

        syncNodeAttrs = {
            "sid":      self.sid,
            "index":    str(self.index),
            "last":     "true" if self.last else "false"
        }

        syncNode = ProtocolTreeNode("sync", syncNodeAttrs)

        node = super(SyncIqProtocolEntity, self).toProtocolTreeNode()
        node.addChild(syncNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        syncNode         = node.getChild("sync")
        entity           = IqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = SyncIqProtocolEntity


        entity.setSyncProps(
            syncNode.getAttributeValue("sid"),
            syncNode.getAttributeValue("index"),
            syncNode.getAttributeValue("last")
            )


        return entity



from yowsup.layers.protocol_messages.protocolentities.test_message_text import TextMessageProtocolEntityTest
from yowsup.layers.protocol_messages.protocolentities.message_text_broadcast import BroadcastTextMessage
from yowsup.structs import ProtocolTreeNode
class BroadcastTextMessageTest(TextMessageProtocolEntityTest):
    def setUp(self):
        super(BroadcastTextMessageTest, self).setUp()
        self.ProtocolEntity = BroadcastTextMessage
        broadcastNode = ProtocolTreeNode("broadcast")
        jids = ["jid1", "jid2"]
        toNodes = [ProtocolTreeNode("to", {"jid" : jid}) for jid in jids]
        broadcastNode.addChildren(toNodes)
        self.node.addChild(broadcastNode)



from yowsup.layers.protocol_chatstate.protocolentities.chatstate_incoming import IncomingChatstateProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

entity = IncomingChatstateProtocolEntity(IncomingChatstateProtocolEntity.STATE_TYPING, "jid@s.whatsapp.net")

class IncomingChatstateProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = IncomingChatstateProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.layers.protocol_groups.protocolentities.iq_groups_list import ListGroupsIqProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

entity = ListGroupsIqProtocolEntity()

class ListGroupsIqProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = ListGroupsIqProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.common import YowConstants
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_iq.protocolentities import ResultIqProtocolEntity


class SuccessLeaveGroupsIqProtocolEntity(ResultIqProtocolEntity):
    '''
    <iq type="result" from="g.us" id="{{ID}}">
        <leave>
            <group id="{{GROUP_JID}}"></group>
        </leave>
    </iq>
    '''

    def __init__(self, _id, groupId):
        super(SuccessLeaveGroupsIqProtocolEntity, self).\
            __init__(_from=YowConstants.WHATSAPP_GROUP_SERVER, _id=_id)
        self.setProps(groupId)

    def setProps(self, groupId):
        self.groupId = groupId

    def __str__(self):
        out = super(SuccessLeaveGroupsIqProtocolEntity, self).__str__()
        out += "Group Id: %s\n" % self.groupId
        return out

    def toProtocolTreeNode(self):
        node = super(SuccessLeaveGroupsIqProtocolEntity, self).\
            toProtocolTreeNode()
        leaveNode = ProtocolTreeNode(
            "leave", {}, [ProtocolTreeNode("group", {"id": self.groupId})]
        )
        node.addChild(leaveNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(SuccessLeaveGroupsIqProtocolEntity, SuccessLeaveGroupsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = SuccessLeaveGroupsIqProtocolEntity
        entity.setProps(
            node.getChild("leave").getChild("group").getAttributeValue("id")
        )
        return entity



from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from yowsup.layers.protocol_iq.protocolentities import ErrorIqProtocolEntity
from yowsup.layers.protocol_iq.protocolentities.iq_result import ResultIqProtocolEntity
from .protocolentities import *
import logging
logger = logging.getLogger(__name__)


class YowGroupsProtocolLayer(YowProtocolLayer):

    HANDLE = (
        CreateGroupsIqProtocolEntity,
        InfoGroupsIqProtocolEntity,
        LeaveGroupsIqProtocolEntity,
        ListGroupsIqProtocolEntity,
        SubjectGroupsIqProtocolEntity,
        ParticipantsGroupsIqProtocolEntity,
        AddParticipantsIqProtocolEntity,
        PromoteParticipantsIqProtocolEntity,
        DemoteParticipantsIqProtocolEntity,
        RemoveParticipantsIqProtocolEntity
    )

    def __init__(self):
        handleMap = {
            "iq": (None, self.sendIq),
            "notification": (self.recvNotification, None)
        }
        super(YowGroupsProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Groups Iq Layer"

    def sendIq(self, entity):
        if entity.__class__ in self.__class__.HANDLE:
            if entity.__class__ == SubjectGroupsIqProtocolEntity:
                self._sendIq(entity, self.onSetSubjectSuccess, self.onSetSubjectFailed)
            elif entity.__class__ == CreateGroupsIqProtocolEntity:
                self._sendIq(entity, self.onCreateGroupSuccess, self.onCreateGroupFailed)
            elif entity.__class__ == ParticipantsGroupsIqProtocolEntity:
                self._sendIq(entity, self.onGetParticipantsResult)
            elif entity.__class__ == AddParticipantsIqProtocolEntity:
                self._sendIq(entity, self.onAddParticipantsSuccess, self.onAddParticipantsFailed)
            elif entity.__class__ == PromoteParticipantsIqProtocolEntity:
                self._sendIq(entity, self.onPromoteParticipantsSuccess, self.onPromoteParticipantsFailed)
            elif entity.__class__ == DemoteParticipantsIqProtocolEntity:
                self._sendIq(entity, self.onDemoteParticipantsSuccess, self.onDemoteParticipantsFailed)
            elif entity.__class__ == RemoveParticipantsIqProtocolEntity:
                self._sendIq(entity, self.onRemoveParticipantsSuccess, self.onRemoveParticipantsFailed)
            elif entity.__class__ == ListGroupsIqProtocolEntity:
                self._sendIq(entity, self.onListGroupsResult)
            elif entity.__class__ == LeaveGroupsIqProtocolEntity:
                self._sendIq(entity, self.onLeaveGroupSuccess, self.onLeaveGroupFailed)
            elif entity.__class__ == InfoGroupsIqProtocolEntity:
                self._sendIq(entity, self.onInfoGroupSuccess, self.onInfoGroupFailed)
            else:
                self.entityToLower(entity)

    def onCreateGroupSuccess(self, node, originalIqEntity):
        logger.info("Group create success")
        self.toUpper(SuccessCreateGroupsIqProtocolEntity.fromProtocolTreeNode(node))

    def onCreateGroupFailed(self, node, originalIqEntity):
        logger.error("Group create failed")
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(node))

    def onSetSubjectSuccess(self, node, originalIqEntity):
        logger.info("Group subject change success")
        self.toUpper(ResultIqProtocolEntity.fromProtocolTreeNode(node))

    def onSetSubjectFailed(self, node, originalIqEntity):
        logger.error("Group subject change failed")
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(node))

    def onGetParticipantsResult(self, node, originalIqEntity):
        self.toUpper(ListParticipantsResultIqProtocolEntity.fromProtocolTreeNode(node))

    def onAddParticipantsSuccess(self, node, originalIqEntity):
        logger.info("Group add participants success")
        self.toUpper(SuccessAddParticipantsIqProtocolEntity.fromProtocolTreeNode(node))

    def onRemoveParticipantsFailed(self, node, originalIqEntity):
        logger.error("Group remove participants failed")
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(node))

    def onRemoveParticipantsSuccess(self, node, originalIqEntity):
        logger.info("Group remove participants success")
        self.toUpper(SuccessRemoveParticipantsIqProtocolEntity.fromProtocolTreeNode(node))

    def onPromoteParticipantsFailed(self, node, originalIqEntity):
        logger.error("Group promote participants failed")
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(node))

    def onPromoteParticipantsSuccess(self, node, originalIqEntity):
        logger.info("Group promote participants success")
        self.toUpper(ResultIqProtocolEntity.fromProtocolTreeNode(node))

    def onDemoteParticipantsFailed(self, node, originalIqEntity):
        logger.error("Group demote participants failed")
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(node))

    def onDemoteParticipantsSuccess(self, node, originalIqEntity):
        logger.info("Group demote participants success")
        self.toUpper(ResultIqProtocolEntity.fromProtocolTreeNode(node))

    def onAddParticipantsFailed(self, node, originalIqEntity):
        logger.error("Group add participants failed")
        self.toUpper(FailureAddParticipantsIqProtocolEntity.fromProtocolTreeNode(node))

    def onListGroupsResult(self, node, originalIqEntity):
        self.toUpper(ListGroupsResultIqProtocolEntity.fromProtocolTreeNode(node))

    def onLeaveGroupSuccess(self, node, originalIqEntity):
        logger.info("Group leave success")
        self.toUpper(SuccessLeaveGroupsIqProtocolEntity.fromProtocolTreeNode(node))

    def onLeaveGroupFailed(self, node, originalIqEntity):
        logger.error("Group leave failed")
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(node))

    def onInfoGroupSuccess(self, node, originalIqEntity):
        logger.info("Group info success")
        self.toUpper(InfoGroupsResultIqProtocolEntity.fromProtocolTreeNode(node))

    def onInfoGroupFailed(self, node, originalIqEntity):
        logger.error("Group info failed")
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(node))

    def recvNotification(self, node):
        if node["type"] == "w:gp2":
            if node.getChild("subject"):
                self.toUpper(SubjectGroupsNotificationProtocolEntity.fromProtocolTreeNode(node))
            elif node.getChild("create"):
                self.toUpper(CreateGroupsNotificationProtocolEntity.fromProtocolTreeNode(node))
            elif node.getChild("remove"):
                self.toUpper(RemoveGroupsNotificationProtocolEntity.fromProtocolTreeNode(node))
            elif node.getChild("add"):
                self.toUpper(AddGroupsNotificationProtocolEntity.fromProtocolTreeNode(node))



import unittest
from yowsup.layers.auth.keystream import KeyStream
class KeyStreamTest(unittest.TestCase):

    def setUp(self):

        self.password = bytearray(list(map(ord,"password")))
        self.salt     = bytearray(list(map(ord,"salt")))

        self.keysTarget = [
            bytearray(b'>\xd5\x8a\xecB\xdb\xc8\xd4}\x98\x9aBa\x89\x9fC\x08\xdcp\x8d'),
            bytearray(b'\xd3;\xda:\x8f\x94CX\xe4;\xbb\xcc"W\x83\xe0m\xba\xe0\xd1'),
            bytearray(b'\x92w{\xc2\x04~\x08;\x81w\xb7h3\xb8\xc4t\xbd\xed\xf7q'),
            bytearray(b'.\xc7\xe4\x15\x1a\xa0\xfd\x98\xc0\xea\xefs{\r7E\xa6\xec\xd5\xfb')
        ]

        self.pbkdbf2_res = bytearray(b'\xeal\x01M\xc7-o\x8c\xcd\x1e\xd9*\xce\x1dA\xf0\xd8\xde\x89W')
        self.inputMessage = bytearray([0,0,0,0, ord('t'), ord('a'), ord('r'), ord('e'), ord('k')])
        self.inputMessageComputedMac = bytearray(b'\xec\x82q\xab\x9f\x8d;\xac\x83\xc5X\xbb\xb6u\x1c\xb0\xd2\x82=`')
        self.inputMessageRC4Ciphered = bytearray(b'\x00\x00\x00\x00\xf4&-\x0c\xbc')


    def test_generateKeys(self):
        keys = KeyStream.generateKeys(self.password, self.salt)
        self.assertEqual(keys, self.keysTarget)

    def test_pbkdf2(self):
        result = KeyStream.pbkdf2(self.password, self.salt, 2, 20)
        result = bytearray(result)
        self.assertEqual(result, self.pbkdbf2_res)
    
    def test_computemac(self):
        keys = self.keysTarget
        kstream = KeyStream(keys[2], keys[3])
        res = kstream.computeMac(self.inputMessage, 4, len(self.inputMessage) - 4)
        self.assertEqual(res, self.inputMessageComputedMac)

    def test_rc4(self):
        keys = self.keysTarget
        kstream = KeyStream(keys[2], keys[3])
        kstream.rc4.cipher(self.inputMessage, 4, len(self.inputMessage) - 4)
        self.assertEqual(self.inputMessage, self.inputMessageRC4Ciphered)

    def test_encodeMessage(self):
        keys = self.keysTarget
        kstream = KeyStream(keys[2], keys[3])
        encoded = kstream.encodeMessage(self.inputMessage, 0, 4, len(self.inputMessage) - 4)

    def test_manyEncode(self):
        keys = self.keysTarget
        kstream = KeyStream(keys[2], keys[3])
        for i in range(0, 300):
            encoded = kstream.encodeMessage(self.inputMessage, 0, 4, len(self.inputMessage) - 4)


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq import IqProtocolEntity
class ResultIqProtocolEntity(IqProtocolEntity):

    '''
    <iq type="result" id="{{id}}" from="{{FROM}}">
    </iq>
    '''

    def __init__(self, xmlns = None, _id = None, to = None, _from = None):
        super(ResultIqProtocolEntity, self).__init__(xmlns = xmlns, _id = _id, _type = "result", to = to, _from = _from)




from yowsup.layers.protocol_acks.protocolentities.ack_outgoing import OutgoingAckProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

entity = OutgoingAckProtocolEntity("12345", "receipt", "delivery", "to_jid")

class OutgoingAckProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = OutgoingAckProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest
class ResultIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(ResultIqProtocolEntityTest, self).setUp()
        self.node.setAttribute("type", "result")


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .ib import IbProtocolEntity
class OfflineIbProtocolEntity(IbProtocolEntity):
    '''
    <ib from="s.whatsapp.net">
        <offline count="{{X}}"></offline>
    </ib>
    '''
    def __init__(self, count):
        super(IbProtocolEntity, self).__init__()
        self.setProps(count)


    def setProps(self, count):
        self.count = int(count)
    
    def toProtocolTreeNode(self):
        node = super(OfflineIbProtocolEntity, self).toProtocolTreeNode()
        offlineChild = ProtocolTreeNode("offline", {"count": str(self.count)})
        node.addChild(offlineChild)
        return node

    def __str__(self):
        out = super(OfflineIbProtocolEntity, self).__str__()
        out += "Offline count: %s\n" % self.count
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = IbProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = OfflineIbProtocolEntity
        entity.setProps(node.getChild("offline")["count"])
        return entity



from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest
from yowsup.layers.protocol_profiles.protocolentities import UnregisterIqProtocolEntity
from yowsup.structs import ProtocolTreeNode

class UnregisterIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(UnregisterIqProtocolEntityTest, self).setUp()
        self.ProtocolEntity = UnregisterIqProtocolEntity
        self.node.addChild(ProtocolTreeNode("remove", {"xmlns": "urn:xmpp:whatsapp:account"}))


from yowsup.layers.interface                           import YowInterfaceLayer, ProtocolEntityCallback

class EchoLayer(YowInterfaceLayer):

    @ProtocolEntityCallback("message")
    def onMessage(self, messageProtocolEntity):

        if messageProtocolEntity.getType() == 'text':
            self.onTextMessage(messageProtocolEntity)
        elif messageProtocolEntity.getType() == 'media':
            self.onMediaMessage(messageProtocolEntity)

        self.toLower(messageProtocolEntity.forward(messageProtocolEntity.getFrom()))
        self.toLower(messageProtocolEntity.ack())
        self.toLower(messageProtocolEntity.ack(True))


    @ProtocolEntityCallback("receipt")
    def onReceipt(self, entity):
        self.toLower(entity.ack())

    def onTextMessage(self,messageProtocolEntity):
        # just print info
        print("Echoing %s to %s" % (messageProtocolEntity.getBody(), messageProtocolEntity.getFrom(False)))

    def onMediaMessage(self, messageProtocolEntity):
        # just print info
        if messageProtocolEntity.getMediaType() == "image":
            print("Echoing image %s to %s" % (messageProtocolEntity.url, messageProtocolEntity.getFrom(False)))

        elif messageProtocolEntity.getMediaType() == "location":
            print("Echoing location (%s, %s) to %s" % (messageProtocolEntity.getLatitude(), messageProtocolEntity.getLongitude(), messageProtocolEntity.getFrom(False)))

        elif messageProtocolEntity.getMediaType() == "vcard":
            print("Echoing vcard (%s, %s) to %s" % (messageProtocolEntity.getName(), messageProtocolEntity.getCardData(), messageProtocolEntity.getFrom(False)))



from .iq import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class PropsIqProtocolEntity(IqProtocolEntity):
    def __init__(self):
        super(PropsIqProtocolEntity, self).__init__("w", _type="get")

    def toProtocolTreeNode(self):
        node = super(PropsIqProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode("props"))
        return node


from .iq import IqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class CryptoIqProtocolEntity(IqProtocolEntity):
    def __init__(self):
        super(CryptoIqProtocolEntity, self).__init__("urn:xmpp:whatsapp:account", _type="get")

    def toProtocolTreeNode(self):
        node = super(CryptoIqProtocolEntity, self).toProtocolTreeNode()
        cryptoNode = ProtocolTreeNode("crypto", {"action": "create"})
        googleNode = ProtocolTreeNode("google", data = "fe5cf90c511fb899781bbed754577098e460d048312c8b36c11c91ca4b49ca34".decode('hex'))
        cryptoNode.addChild(googleNode)
        node.addChild(cryptoNode)
        return node


from yowsup.layers.protocol_contacts.protocolentities import UpdateContactNotificationProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import time
import unittest

entity = UpdateContactNotificationProtocolEntity("1234", "jid@s.whatsapp.net",
                                                 int(time.time()), "notify", False,"contactjid@s.whatsapp.net")

class UpdateContactNotificationProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = UpdateContactNotificationProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from yowsup.layers.protocol_notifications.protocolentities import NotificationProtocolEntity
from .notification_groups import GroupsNotificationProtocolEntity

class SubjectGroupsNotificationProtocolEntity(GroupsNotificationProtocolEntity):
    '''

    <notification notify="WhatsApp" id="{{id}}" t="{{TIMESTAMP}}" participant="{{PARTICIPANT_JID}}" from="{{GROUP_JID}}" type="w:gp2">
        <subject s_t="{{subject_set_timestamp}}" s_o="{{subject_owner_jid}}" subject="{{SUBJECT}}">
        </subject>
    </notification>

    '''

    def __init__(self, _type, _id,  _from, timestamp, notify, participant, subject):
        super(SubjectGroupsNotificationProtocolEntity, self).__init__(_id, _from, timestamp, notify, participant)
        self.setSubjectData(subject)

    def setSubjectData(self, subject, subjectOwner, subjectTimestamp):
        self.subject = subject
        self.subjectOwner = subjectOwner
        self.subjectTimestamp = int(subjectTimestamp)

    def getSubject(self):
        return self.subject

    def getSubjectOwner(self, full = True):
        return self.subjectOwner if full else self.subjectOwner.split('@')[0]

    def getSubjectTimestamp(self):
        return self.subjectTimestamp

    def __str__(self):
        out = super(SubjectGroupsNotificationProtocolEntity, self).__str__()
        out += "New subject: %s\n" % self.getSubject()
        out += "Set by: %s\n" % self.getSubjectOwner()
        return out

    def toProtocolTreeNode(self):
        node = super(SubjectGroupsNotificationProtocolEntity, self).toProtocolTreeNode()
        subjectNode = ProtocolTreeNode("subject", {
            "s_t": str(self.getSubjectTimestamp()),
            "s_o": self.getSubjectOwner(),
            "subject": self.getSubject()
        })
        node.addChild(subjectNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(SubjectGroupsNotificationProtocolEntity, SubjectGroupsNotificationProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = SubjectGroupsNotificationProtocolEntity
        subjectNode = node.getChild("subject")
        entity.setSubjectData(subjectNode["subject"], subjectNode["s_o"], subjectNode["s_t"])
        return entity



from yowsup.layers.protocol_ib.protocolentities.ib import IbProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class IbProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = IbProtocolEntity
        self.node = ProtocolTreeNode("ib")



from yowsup.layers.protocol_receipts.protocolentities import IncomingReceiptProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest
import time

class IncomingReceiptProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = IncomingReceiptProtocolEntity
        self.node = IncomingReceiptProtocolEntity("123", "sender", int(time.time())).toProtocolTreeNode()



from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest
from yowsup.layers.protocol_profiles.protocolentities import GetPrivacyIqProtocolEntity
from yowsup.structs import ProtocolTreeNode

entity = GetPrivacyIqProtocolEntity()

class GetPrivacyIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(GetPrivacyIqProtocolEntityTest, self).setUp()
        self.ProtocolEntity = GetPrivacyIqProtocolEntity
        self.node = entity.toProtocolTreeNode()



from .notification_groups import GroupsNotificationProtocolEntity
from yowsup.structs import ProtocolTreeNode
class RemoveGroupsNotificationProtocolEntity(GroupsNotificationProtocolEntity):
    '''
<notification notify="{{NOTIFY_NAME}}" id="{{id}}" t="{{TIMESTAMP}}" participant="{{participant_jiid}}" from="{{group_jid}}" type="w:gp2">
<remove subject="{{subject}}">
<participant jid="{{participant_jid}}">
</participant>
</remove>
</notification>
    '''
    TYPE_PARTICIPANT_ADMIN = "admin"
    def __init__(self, _id, _from, timestamp, notify, participant, offline,
                 subject,
                 participants):
        super(RemoveGroupsNotificationProtocolEntity, self).__init__(_id, _from, timestamp, notify, participant, offline)
        self.setGroupProps(subject, participants)

    def setGroupProps(self,
                      subject,
                      participants):

        assert type(participants) is dict, "Participants must be a dict {jid => type?}"

        self.subject = subject
        self.participants = participants

    def getParticipants(self):
        return self.participants

    def getSubject(self):
        return self.subject

    def toProtocolTreeNode(self):
        node = super(RemoveGroupsNotificationProtocolEntity, self).toProtocolTreeNode()
        removeNode = ProtocolTreeNode("remove", {"subject": self.subject})
        participants = []
        for jid in self.getParticipants():
            pnode = ProtocolTreeNode("participant", {"jid": jid})
            participants.append(pnode)

        removeNode.addChildren(participants)
        node.addChild(removeNode)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        removeNode = node.getChild("remove")
        participants = {}
        for p in removeNode.getAllChildren("participant"):
            participants[p["jid"]] = p["type"]

        return RemoveGroupsNotificationProtocolEntity(
            node["id"], node["from"], node["t"], node["notify"], node["participant"], node["offline"],
            removeNode["subject"], participants
        )



from yowsup.structs import ProtocolTreeNode
from .notification_contact import ContactNotificationProtocolEntity
class AddContactNotificationProtocolEntity(ContactNotificationProtocolEntity):
    '''
    <notification offline="0" id="{{NOTIFICATION_ID}}" notify="{{NOTIFY_NAME}}" type="contacts" 
            t="{{TIMESTAMP}}" from="{{SENDER_JID}}">
        <add jid="{{SET_JID}}"> </add>
    </notification>
    '''

    def __init__(self, _id,  _from, timestamp, notify, offline, contactJid):
        super(AddContactNotificationProtocolEntity, self).__init__(_id, _from, timestamp, notify, offline)
        self.setData(contactJid)

    def setData(self, jid):
        self.contactJid = jid

    def toProtocolTreeNode(self):
        node = super(AddContactNotificationProtocolEntity, self).toProtocolTreeNode()
        removeNode = ProtocolTreeNode("add", {"jid": self.contactJid}, None, None)
        node.addChild(removeNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = ContactNotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = AddContactNotificationProtocolEntity
        removeNode = node.getChild("add")
        entity.setData(removeNode.getAttributeValue("jid"))
        return entity


import hashlib, hmac, sys
from struct import pack
from operator import xor
from itertools import starmap

class RC4:
    def __init__(self, key, drop):
        self.s = []
        self.i = 0;
        self.j = 0;

        self.s = [0] * 256

        for i in range(0, len(self.s)):
            self.s[i] = i

        for i in range(0, len(self.s)):
            self.j = (self.j + self.s[i] + key[i % len(key)]) % 256
            RC4.swap(self.s, i, self.j)

        self.j = 0;
        self.cipher(bytearray(drop), 0, drop)


    def cipher(self, data, offset, length):
        while True:
            num = length
            length = num - 1

            if num == 0: break

            self.i = (self.i+1) % 256
            self.j = (self.j + self.s[self.i]) % 256

            RC4.swap(self.s, self.i, self.j)

            num2 = offset
            offset = num2 + 1

            data[num2] = (data[num2] ^ self.s[(self.s[self.i] + self.s[self.j]) % 256])

    @staticmethod
    def swap(arr, i, j):
        tmp = arr[i]
        arr[i] = arr[j]
        arr[j] = tmp


class KeyStream:

    def __init__(self, key, macKey):
        self.key = key if sys.version_info < (3, 0) else bytes(key)
        self.rc4 = RC4(self.key, 0x300)
        self.macKey = str(macKey) if sys.version_info < (3, 0) else bytes(macKey)
        self.seq = 0

    def computeMac(self, bytes_buffer, int_offset, int_length):
        mac = hmac.new(self.macKey, None, hashlib.sha1)
        updateData = bytes_buffer[int_offset:] + bytearray([self.seq >> 24, (self.seq >> 16) % 256, (self.seq >> 8) % 256, self.seq % 256])

        try:
            mac.update(updateData)
        except TypeError: #python3 support
            mac.update(bytes(updateData))

        self.seq += 1
        return bytearray(mac.digest())

    def decodeMessage(self, bufdata, macOffset, offset, length):
        buf = bufdata[:-4]
        hashed = bufdata[-4:]
        numArray = self.computeMac(buf, 0, len(buf))

        num = 0
        while num < 4:
            if numArray[macOffset + num] == hashed[num]:
                num += 1
            else:
                raise Exception("INVALID MAC")

        self.rc4.cipher(buf, 0, len(buf))

        return buf

    def encodeMessage(self, buf, macOffset, offset, length):
        self.rc4.cipher(buf, offset, length)
        mac = self.computeMac(buf, offset, length)
        output = buf[0:macOffset] + mac[0:4] + buf[macOffset+4:]
        return output

    @staticmethod
    def generateKeys(password, nonce):
        resultBytes = []

        for i in range(1, 5):
            currNonce = nonce + bytearray([i])
            resultBytes.append(KeyStream.pbkdf2(password, currNonce, 2, 20))

        return resultBytes

    #@staticmethod  ##use if drop python-2.6 support
    #def pbkdf2( password, salt, itercount, keylen):
    #    return bytearray(hashlib.pbkdf2_hmac('sha1', password, salt, itercount, keylen))

    @staticmethod
    def pbkdf2( password, salt, itercount, keylen, hashfn = hashlib.sha1 ):
        def pbkdf2_F( h, salt, itercount, blocknum ):
            def prf( h, data ):
                hm = h.copy()
                try:
                    hm.update(bytearray(data))
                except TypeError: #python 3 support
                    hm.update(bytes(data))

                d = hm.digest()
                return bytearray(d)

            U = prf( h, salt + pack('>i',blocknum ) )
            T = U

            for i in range(2, itercount + 1):
                U = prf( h, U )
                T = starmap(xor, zip(T, U))
            return T

        digest_size = hashfn().digest_size
        l = int(keylen / digest_size)
        if keylen % digest_size != 0:
            l += 1
        
        h = hmac.new(bytes(password), None, hashfn )

        T = bytearray()
        for i in range(1, l+1):
            tmp = pbkdf2_F( h, salt, itercount, i )
            T.extend(tmp)
        
        return T[0: keylen]


from yowsup.layers import YowLayer, YowLayerEvent, YowProtocolLayer
from .protocolentities import *
from yowsup.layers.protocol_acks.protocolentities import OutgoingAckProtocolEntity
class YowNotificationsProtocolLayer(YowProtocolLayer):

    def __init__(self):
        handleMap = {
            "notification": (self.recvNotification, self.sendNotification)
        }
        super(YowNotificationsProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "notification Ib Layer"

    def sendNotification(self, entity):
        if entity.getTag() == "notification":
            self.toLower(entity.toProtocolTreeNode())

    def recvNotification(self, node):
        if node["type"] == "picture":
            if node.getChild("set"):
                self.toUpper(SetPictureNotificationProtocolEntity.fromProtocolTreeNode(node))
            elif node.getChild("delete"):
                self.toUpper(DeletePictureNotificationProtocolEntity.fromProtocolTreeNode(node))
            else:
                self.raiseErrorForNode(node)
        elif node["type"] == "status":
            self.toUpper(StatusNotificationProtocolEntity.fromProtocolTreeNode(node))
        elif node["type"] == "features":
            # Not implemented
            pass
        elif node["type"] in [ "contacts", "subject", "w:gp2" ]:
            # Implemented in respectively the protocol_contacts and protocol_groups layer
            pass
        elif node["type"] == "contacts":
            pass
        elif node["type"] == "web":
            # Not implemented
            pass
        else:
            self.raiseErrorForNode(node)

        ack = OutgoingAckProtocolEntity(node["id"], "notification", node["type"], node["from"])
        self.toLower(ack.toProtocolTreeNode())









class TokenDictionary:

    def __init__(self):
        self.dictionary = [
            "",
            "",
            "",
            "account",
            "ack",
            "action",
            "active",
            "add",
            "after",
            "all",
            "allow",
            "apple",
            "auth",
            "author",
            "available",
            "bad-protocol",
            "bad-request",
            "before",
            "body",
            "broadcast",
            "cancel",
            "category",
            "challenge",
            "chat",
            "clean",
            "code",
            "composing",
            "config",
            "contacts",
            "count",
            "create",
            "creation",
            "debug",
            "default",
            "delete",
            "delivery",
            "delta",
            "deny",
            "digest",
            "dirty",
            "duplicate",
            "elapsed",
            "enable",
            "encoding",
            "error",
            "event",
            "expiration",
            "expired",
            "fail",
            "failure",
            "false",
            "favorites",
            "feature",
            "features",
            "feature-not-implemented",
            "field",
            "first",
            "free",
            "from",
            "g.us",
            "get",
            "google",
            "group",
            "groups",
            "groups_v2",
            "http://etherx.jabber.org/streams",
            "http://jabber.org/protocol/chatstates",
            "ib",
            "id",
            "image",
            "img",
            "index",
            "internal-server-error",
            "ip",
            "iq",
            "item-not-found",
            "item",
            "jabber:iq:last",
            "jabber:iq:privacy",
            "jabber:x:event",
            "jid",
            "kind",
            "last",
            "leave",
            "list",
            "max",
            "mechanism",
            "media",
            "message_acks",
            "message",
            "method",
            "microsoft",
            "missing",
            "modify",
            "mute",
            "name",
            "nokia",
            "none",
            "not-acceptable",
            "not-allowed",
            "not-authorized",
            "notification",
            "notify",
            "off",
            "offline",
            "order",
            "owner",
            "owning",
            "p_o",
            "p_t",
            "paid",
            "participant",
            "participants",
            "participating",
            "paused",
            "picture",
            "pin",
            "ping",
            "platform",
            "port",
            "presence",
            "preview",
            "probe",
            "prop",
            "props",
            "query",
            "raw",
            "read",
            "readreceipts",
            "reason",
            "receipt",
            "relay",
            "remote-server-timeout",
            "remove",
            "request",
            "required",
            "resource-constraint",
            "resource",
            "response",
            "result",
            "retry",
            "rim",
            "s_o",
            "s_t",
            "s.us",
            "s.whatsapp.net",
            "seconds",
            "server-error",
            "server",
            "service-unavailable",
            "set",
            "show",
            "silent",
            "stat",
            "status",
            "stream:error",
            "stream:features",
            "subject",
            "subscribe",
            "success",
            "sync",
            "t",
            "text",
            "timeout",
            "timestamp",
            "to",
            "true",
            "type",
            "unavailable",
            "unsubscribe",
            "uri",
            "url",
            "urn:ietf:params:xml:ns:xmpp-sasl",
            "urn:ietf:params:xml:ns:xmpp-stanzas",
            "urn:ietf:params:xml:ns:xmpp-streams",
            "urn:xmpp:ping",
            "urn:xmpp:whatsapp:account",
            "urn:xmpp:whatsapp:dirty",
            "urn:xmpp:whatsapp:mms",
            "urn:xmpp:whatsapp:push",
            "urn:xmpp:whatsapp",
            "user",
            "user-not-found",
            "value",
            "version",
            "w:g",
            "w:p:r",
            "w:p",
            "w:profile:picture",
            "w",
            "wait",
            "WAUTH-2",
            "xmlns:stream",
            "xmlns",
            "1",
            "chatstate",
            "crypto",
            "phash",
            "enc",
            "class",
            "off_cnt",
            "w:g2",
            "promote",
            "demote",
            "creator",
            "Bell.caf",
            "Boing.caf",
            "Glass.caf",
            "Harp.caf",
            "TimePassing.caf",
            "Tri-tone.caf",
            "Xylophone.caf",
            "background",
            "backoff",
            "chunked",
            "context",
            "full",
            "in",
            "interactive",
            "out",
            "registration",
            "sid",
            "urn:xmpp:whatsapp:sync",
            "flt",
            "s16",
            "u8",
            "adpcm",
            "amrnb",
            "amrwb",
            "mp3",
            "pcm",
            "qcelp",
            "wma",
            "h263",
            "h264",
            "jpeg"
        ]

        self.secondaryDictionary = [
            "mpeg4",
            "wmv",
            "audio/3gpp",
            "audio/aac",
            "audio/amr",
            "audio/mp4",
            "audio/mpeg",
            "audio/ogg",
            "audio/qcelp",
            "audio/wav",
            "audio/webm",
            "audio/x-caf",
            "audio/x-ms-wma",
            "image/gif",
            "image/jpeg",
            "image/png",
            "video/3gpp",
            "video/avi",
            "video/mp4",
            "video/mpeg",
            "video/quicktime",
            "video/x-flv",
            "video/x-ms-asf",
            "302",
            "400",
            "401",
            "402",
            "403",
            "404",
            "405",
            "406",
            "407",
            "409",
            "410",
            "500",
            "501",
            "503",
            "504",
            "abitrate",
            "acodec",
            "app_uptime",
            "asampfmt",
            "asampfreq",
            "audio",
            "clear",
            "conflict",
            "conn_no_nna",
            "cost",
            "currency",
            "duration",
            "extend",
            "file",
            "fps",
            "g_notify",
            "g_sound",
            "gcm",
            "gone",
            "google_play",
            "hash",
            "height",
            "invalid",
            "jid-malformed",
            "latitude",
            "lc",
            "lg",
            "live",
            "location",
            "log",
            "longitude",
            "max_groups",
            "max_participants",
            "max_subject",
            "mimetype",
            "mode",
            "napi_version",
            "normalize",
            "orighash",
            "origin",
            "passive",
            "password",
            "played",
            "policy-violation",
            "pop_mean_time",
            "pop_plus_minus",
            "price",
            "pricing",
            "redeem",
            "Replaced by new connection",
            "resume",
            "signature",
            "size",
            "sound",
            "source",
            "system-shutdown",
            "username",
            "vbitrate",
            "vcard",
            "vcodec",
            "video",
            "width",
            "xml-not-well-formed",
            "checkmarks",
            "image_max_edge",
            "image_max_kbytes",
            "image_quality",
            "ka",
            "ka_grow",
            "ka_shrink",
            "newmedia",
            "library",
            "caption",
            "forward",
            "c0",
            "c1",
            "c2",
            "c3",
            "clock_skew",
            "cts",
            "k0",
            "k1",
            "login_rtt",
            "m_id",
            "nna_msg_rtt",
            "nna_no_off_count",
            "nna_offline_ratio",
            "nna_push_rtt",
            "no_nna_con_count",
            "off_msg_rtt",
            "on_msg_rtt",
            "stat_name",
            "sts",
            "suspect_conn",
            "lists",
            "self",
            "qr",
            "web",
            "w:b",
            "recipient",
            "w:stats",
            "forbidden",
            "aurora.m4r",
            "bamboo.m4r",
            "chord.m4r",
            "circles.m4r",
            "complete.m4r",
            "hello.m4r",
            "input.m4r",
            "keys.m4r",
            "note.m4r",
            "popcorn.m4r",
            "pulse.m4r",
            "synth.m4r",
            "filehash",
            "max_list_recipients",
            "en-AU",
            "en-GB",
            "es-MX",
            "pt-PT",
            "zh-Hans",
            "zh-Hant",
            "relayelection",
            "relaylatency",
            "interruption",
            "Apex.m4r",
            "Beacon.m4r",
            "Bulletin.m4r",
            "By The Seaside.m4r",
            "Chimes.m4r",
            "Circuit.m4r",
            "Constellation.m4r",
            "Cosmic.m4r",
            "Crystals.m4r",
            "Hillside.m4r",
            "Illuminate.m4r",
            "Night Owl.m4r",
            "Opening.m4r",
            "Playtime.m4r",
            "Presto.m4r",
            "Radar.m4r",
            "Radiate.m4r",
            "Ripples.m4r",
            "Sencha.m4r",
            "Signal.m4r",
            "Silk.m4r",
            "Slow Rise.m4r",
            "Stargaze.m4r",
            "Summit.m4r",
            "Twinkle.m4r",
            "Uplift.m4r",
            "Waves.m4r",
            "voip",
            "eligible",
            "upgrade",
            "planned",
            "current",
            "future",
            "disable",
            "expire",
            "start",
            "stop",
            "accuracy",
            "speed",
            "bearing",
            "recording",
            "encrypt",
            "key",
            "identity",
            "w:gp2",
            "admin",
            "locked",
            "unlocked",
            "new",
            "battery",
            "archive",
            "adm",
            "plaintext_size",
            "compressed_size",
            "delivered",
            "msg",
            "pkmsg",
            "everyone",
            "v",
            "transport",
            "call-id"
        ]


    def getToken(self, index, secondary = False):
        targetDict = self.dictionary
        if secondary:
            targetDict = self.secondaryDictionary
        elif index > 236 and index < (236 + len(self.secondaryDictionary)):
            targetDict = self.secondaryDictionary
            index = index - 237
        if index < 0 or index > len(targetDict) - 1:
            return None

        return targetDict[index]

    def getIndex(self, token):
        if token in self.dictionary:
            return (self.dictionary.index(token), False)
        elif token in self.secondaryDictionary:
            return (self.secondaryDictionary.index(token), True)

        return None




from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from yowsup.layers.protocol_notifications.protocolentities import NotificationProtocolEntity
class ContactNotificationProtocolEntity(NotificationProtocolEntity):
    '''
    <notification offline="0" id="{{NOTIFICATION_ID}}" notify="{{NOTIFY_NAME}}" type="contacts" 
            t="{{TIMESTAMP}}" from="{{SENDER_JID}}">
    </notification>
    
    '''

    def __init__(self, _id,  _from, timestamp, notify, offline = False):
        super(ContactNotificationProtocolEntity, self).__init__("contacts", _id, _from, timestamp, notify, offline)


    @staticmethod
    def fromProtocolTreeNode(node):
        entity = NotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ContactNotificationProtocolEntity
        return entity


from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_ib.protocolentities.clean_iq import CleanIqProtocolEntity
from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest

class CleanIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(CleanIqProtocolEntityTest, self).setUp()
        self.ProtocolEntity = CleanIqProtocolEntity
        cleanNode = ProtocolTreeNode("clean", {"type": "groups"})
        self.node.addChild(cleanNode)


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class StreamErrorAckProtocolEntity(ProtocolEntity):
    '''
     <stream:error>
        <ack></ack>
     </stream:error>
    '''
    def __init__(self):
        super(StreamErrorAckProtocolEntity, self).__init__("stream:error")

    def toProtocolTreeNode(self):
        node = super(StreamErrorAckProtocolEntity, self).toProtocolTreeNode()
        node.addChild(ProtocolTreeNode("ack"))
        return node

    def __str__(self):
        out = "Ack Stream Error\n"
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        return StreamErrorAckProtocolEntity()



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_picture import PictureIqProtocolEntity
class ListPicturesIqProtocolEntity(PictureIqProtocolEntity):
    '''
    <iq type="get" id="{{id}}" xmlns="w:profile:picture", to="self.jid">
        <list>
            <user jid="{{user_jid}}"></user>
            <user jid="{{user_jid}}"></user>
        </list>
    </iq>
    '''

    def __init__(self, selfJid, jids):
        super(ListPicturesIqProtocolEntity, self).__init__(jid = selfJid, type = "get")
        self.setProps(jids)

    def setProps(self, jids):
        assert type(jids) is list and len(jids), "Must specify a list of jids to get the pictures for"
        self.jids = jids

    def toProtocolTreeNode(self):
        node = super(ListPicturesIqProtocolEntity, self).toProtocolTreeNode()
        userNodes = [ProtocolTreeNode("user", {"jid": jid}) for jid in self.jids]
        listNode = ProtocolTreeNode("list", {}, userNodes)
        node.addChild(listNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = PictureIqProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ListPicturesIqProtocolEntity
        jids = [userNode.getAttributeValue("jid") for userNode in node.getChild("list").getAllChildren()]
        entity.setProps(jids)
        return entity


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .message_media import MediaMessageProtocolEntity

class VCardMediaMessageProtocolEntity(MediaMessageProtocolEntity):
    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}" 
    offline="{{OFFLINE}}" type="text" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">
        <media type="vcard">
            <vcard name="Hany Yasser">
                BEGIN:VCARD
                VERSION:3.0
                N:Yasser;Hany;;;
                FN:Hany Yasser
                PHOTO;BASE64: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
                BDAY;value=date:1989-01-05
                ORG:Vodafone Egypt;
                item1.EMAIL;type=INTERNET:hanyyasser@hotmail.com
                item1.X-ABLabel:INTERNET
                item2.EMAIL;type=INTERNET:hanybotbot@hotmail.com
                item2.X-ABLabel:INTERNET
                item3.ADR;type=HOME:;;Heliopolis;Cairo;Al Qahirah;;Egypt
                item4.ADR;type=HOME:;;;cairo;;;Egypt
                item5.URL:http://www.facebook.com/profile.php?id=626850952
                item5.X-ABLabel:_$!<HomePage>!$_
                X-FACEBOOK:hany.yasser1
                END:VCARD
            </vcard>
        </media>
    </message>
    '''


    def __init__(self, name, card_data, _id = None, _from = None, to = None, notify = None, timestamp = None, participant = None,
            preview = None, offline = None, retry = None):

        super(VCardMediaMessageProtocolEntity, self).__init__("vcard", _id, _from, to, notify, timestamp, participant, preview, offline, retry)
        self.setVcardMediaProps(name,card_data)

    def __str__(self):
        out  = super(MediaMessageProtocolEntity, self).__str__()
        out += "Name: %s\n" % self.name
        out += "Card Data: %s\n" % self.card_data
        return out

    def getName(self):
        return self.name

    def getCardData(self):
        return self.card_data
   
    def setVcardMediaProps(self, name, card_data):
        self.name = name
        self.card_data = card_data

    def toProtocolTreeNode(self):
        node = super(VCardMediaMessageProtocolEntity, self).toProtocolTreeNode()
        mediaNode = node.getChild("media")
        mediaNode["type"] = "vcard"
        vcardNode = ProtocolTreeNode("vcard", {"name":self.name}, None,self.card_data)
        mediaNode.addChild(vcardNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = MediaMessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = VCardMediaMessageProtocolEntity
        mediaNode = node.getChild("media")
        entity.setVcardMediaProps(
            mediaNode.getAllChildren()[0].getAttributeValue('name'),
            mediaNode.getChild("vcard").getData()
        )
        return entity


from yowsup.layers.protocol_iq.protocolentities.test_iq_result import ResultIqProtocolEntityTest
from yowsup.layers.protocol_media.protocolentities import ResultRequestUploadIqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class ResultRequestUploadIqProtocolEntityTest(ResultIqProtocolEntityTest):
    def setUp(self):
        super(ResultRequestUploadIqProtocolEntityTest, self).setUp()
        mediaNode = ProtocolTreeNode("media", {"url": "url", "ip": "1.2.3.4"})
        self.ProtocolEntity = ResultRequestUploadIqProtocolEntity
        self.node.addChild(mediaNode)


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .message_media_downloadable import DownloadableMediaMessageProtocolEntity
class AudioDownloadableMediaMessageProtocolEntity(DownloadableMediaMessageProtocolEntity):
    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}" 
        offline="{{OFFLINE}}" type="text" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">
        <media type="{{DOWNLOADABLE_MEDIA_TYPE: (image | audio | video)}}"
            mimetype="{{MIME_TYPE}}" 
            filehash="{{FILE_HASH}}"
            url="{{DOWNLOAD_URL}}" 
            ip="{{IP}}"
            size="{{MEDIA SIZE}}"
            file="{{FILENAME}}" 


            encoding="{{ENCODING}}" 
            height="{{IMAGE_HEIGHT}}" 
            width="{{IMAGE_WIDTH}}"
            
            > {{THUMBNAIL_RAWDATA (JPEG?)}}
        </media>
    </message>
    '''
    def __init__(self,
            mimeType, fileHash, url, ip, size, fileName, 
            abitrate, acodec, asampfreq, duration, encoding, origin, seconds,
            _id = None, _from = None, to = None, notify = None, timestamp = None, 
            participant = None, preview = None, offline = None, retry = None):

        super(AudioDownloadableMediaMessageProtocolEntity, self).__init__("audio",
            mimeType, fileHash, url, ip, size, fileName,
            _id, _from, to, notify, timestamp, participant, preview, offline, retry)
        self.setAudioProps(abitrate, acodec, asampfreq, duration, encoding, origin, seconds)

    def __str__(self):
        out  = super(AudioDownloadableMediaMessageProtocolEntity, self).__str__()
        out += "Bitrate: %s\n" % self.abitrate
        out += "Codec: %s\n" % self.acodec
        out += "Duration: %s\n" % self.duration
        out += "Encoding: %s\n" % self.encoding
        out += "Origin: %s\n" % self.origin
        out += "Sampling freq.: %s\n" % self.asampfreq
        return out

    def setAudioProps(self, abitrate = None, acodec = None, asampfreq = None,
                      duration = None, encoding = None, origin = None, seconds = None):
        self.abitrate  = abitrate
        self.acodec    = acodec
        self.asampfreq = asampfreq
        self.duration  = duration
        self.encoding  = encoding
        self.origin    = origin
        self.seconds   = seconds

    def toProtocolTreeNode(self):
        node = super(AudioDownloadableMediaMessageProtocolEntity, self).toProtocolTreeNode()
        mediaNode = node.getChild("media")

        if self.abitrate:
            mediaNode.setAttribute("abitrate",  self.abitrate)
        if self.acodec:
            mediaNode.setAttribute("acodec",    self.acodec)
        if self.asampfreq:
            mediaNode.setAttribute("asampfreq", self.asampfreq)
        if self.duration:
            mediaNode.setAttribute("duration",  self.duration)
        if self.encoding:
            mediaNode.setAttribute("encoding",  self.encoding)
        if self.origin:
            mediaNode.setAttribute("origin",    self.origin)
        if self.seconds:
            mediaNode.setAttribute("seconds",   self.seconds)

        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = DownloadableMediaMessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = AudioDownloadableMediaMessageProtocolEntity
        mediaNode = node.getChild("media")
        entity.setAudioProps(
            mediaNode.getAttributeValue("abitrate"),
            mediaNode.getAttributeValue("acodec"),
            mediaNode.getAttributeValue("asampfreq"),
            mediaNode.getAttributeValue("duration"),
            mediaNode.getAttributeValue("encoding"),
            mediaNode.getAttributeValue("origin"),
            mediaNode.getAttributeValue("seconds"),
        )
        return entity



    @staticmethod
    def fromFilePath(fpath, url, ip, to, mimeType = None, preview = None, filehash = None, filesize = None):
        entity = DownloadableMediaMessageProtocolEntity.fromFilePath(fpath, url, DownloadableMediaMessageProtocolEntity.MEDIA_TYPE_AUDIO, ip, to, mimeType, preview)
        entity.__class__ = AudioDownloadableMediaMessageProtocolEntity
        entity.setAudioProps()
        return entity



from yowsup.common import YowConstants
from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_groups import GroupsIqProtocolEntity
class CreateGroupsIqProtocolEntity(GroupsIqProtocolEntity):
    '''
    <iq type="set" id="{{id}}" xmlns="w:g2", to="g.us">
        <create subject="{{subject}}">
             <participant jid="{{jid}}"></participant>
        </create>
    </iq>
    '''

    def __init__(self, subject, _id = None, participants = None):
        super(CreateGroupsIqProtocolEntity, self).__init__(to = YowConstants.WHATSAPP_GROUP_SERVER, _id = _id, _type = "set")
        self.setProps(subject)
        self.setParticipants(participants or [])

    def setProps(self, subject):
        self.subject = subject

    def setParticipants(self, participants):
        self.participantList = participants

    def toProtocolTreeNode(self):
        node = super(CreateGroupsIqProtocolEntity, self).toProtocolTreeNode()
        cnode = ProtocolTreeNode("create",{ "subject": self.subject})
        participantNodes = [
            ProtocolTreeNode("participant", {
                "jid":       participant
            })
            for participant in self.participantList
        ]
        cnode.addChildren(participantNodes)
        node.addChild(cnode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(CreateGroupsIqProtocolEntity,CreateGroupsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = CreateGroupsIqProtocolEntity
        entity.setProps(node.getChild("create").getAttributeValue("subject"))
        participantList = []
        for participantNode in node.getChild("create").getAllChildren():
            participantList.append(participantNode["jid"])
        entity.setParticipants(participantList)
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .message_media import MediaMessageProtocolEntity

class LocationMediaMessageProtocolEntity(MediaMessageProtocolEntity):
    '''
    <message t="{{TIME_STAMP}}" from="{{CONTACT_JID}}" 
    offline="{{OFFLINE}}" type="text" id="{{MESSAGE_ID}}" notify="{{NOTIFY_NAME}}">
        <media 
            latitude="52.52393" 
            type="location"
            longitude="13.41747"
            name="Location Name"
            url="http://www.foursquare.com/XXXX"
            encoding="raw"
        >{{THUMBNAIL_RAWDATA}}</media>
    </message>
    '''


    def __init__(self, latitude, longitude, name, url, encoding, _id = None, _from = None, to = None, notify = None, timestamp = None, participant = None,
            preview = None, offline = None, retry = None):

        super(LocationMediaMessageProtocolEntity, self).__init__("location", _id, _from, to, notify, timestamp, participant, preview, offline, retry)
        self.setLocationMediaProps(latitude,longitude,name,url,encoding)

    def __str__(self):
        out  = super(MediaMessageProtocolEntity, self).__str__()
        out += "Latitude: %s\n" % self.latitude
        out += "Longitude: %s\n" % self.longitude
        out += "Name: %s\n" % self.name
        out += "URL: %s\n" % self.url
        out += "Encoding: %s\n" % self.encoding

        return out

    def getLatitude(self):
        return self.latitude

    def getLongitude(self):
        return self.longitude

    def getLocationName(self):
        return self.name

    def getLocationURL(self):
        return self.url

    def setLocationMediaProps(self, latitude, longitude, locationName, url, encoding):
        self.latitude = str(latitude)
        self.longitude = str(longitude)
        self.name = locationName
        self.url = url
        self.encoding= encoding

    def toProtocolTreeNode(self):
        node = super(LocationMediaMessageProtocolEntity, self).toProtocolTreeNode()
        mediaNode = node.getChild("media")
        mediaNode.setAttribute("latitude",  self.latitude)
        mediaNode.setAttribute("longitude",  self.longitude)
        mediaNode.setAttribute("encoding", self.encoding)

        if self.name:
            mediaNode.setAttribute("name", self.name)
        if self.url:
            mediaNode.setAttribute("url", self.url)
            
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = MediaMessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = LocationMediaMessageProtocolEntity
        mediaNode = node.getChild("media")
        entity.setLocationMediaProps(
            mediaNode.getAttributeValue("latitude"),
            mediaNode.getAttributeValue("longitude"),
            mediaNode.getAttributeValue("name"),
            mediaNode.getAttributeValue("url"),
            mediaNode.getAttributeValue("encoding")
            )
        return entity



import time,datetime,re, hashlib
from dateutil import tz
import os
from .constants import YowConstants
import codecs, sys
import logging
import tempfile
import base64
import hashlib
import os.path, mimetypes
from .optionalmodules import PILOptionalModule, FFVideoOptionalModule

logger = logging.getLogger(__name__)

class Jid:
    @staticmethod
    def normalize(number):
        if '@' in number:
            return number
        elif "-" in number:
            return "%s@%s" % (number, YowConstants.WHATSAPP_GROUP_SERVER)
        return "%s@%s" % (number, YowConstants.WHATSAPP_SERVER)

class HexTools:
    decode_hex = codecs.getdecoder("hex_codec")
    @staticmethod
    def decodeHex(hexString):
        result = HexTools.decode_hex(hexString)[0]
        if sys.version_info >= (3,0):
            result = result.decode('latin-1')
        return result

class WATools:
    @staticmethod
    def generateIdentity():
        return os.urandom(20)

    @staticmethod
    def getFileHashForUpload(filePath):
        sha1 = hashlib.sha256()
        f = open(filePath, 'rb')
        try:
            sha1.update(f.read())
        finally:
            f.close()
        b64Hash = base64.b64encode(sha1.digest())
        return b64Hash if type(b64Hash) is str else b64Hash.decode()

class StorageTools:
    @staticmethod
    def constructPath(*path):
        path = os.path.join(*path)
        fullPath = os.path.expanduser(os.path.join(YowConstants.PATH_STORAGE, path))
        if not os.path.exists(os.path.dirname(fullPath)):
            os.makedirs(os.path.dirname(fullPath))
        return fullPath

    @staticmethod
    def getStorageForPhone(phone):
        return StorageTools.constructPath(phone + '/')

    @staticmethod
    def writeIdentity(phone, identity):
        path = StorageTools.getStorageForPhone(phone)
        with open(os.path.join(path, "id"), 'wb') as idFile:
            idFile.write(identity)

    @staticmethod
    def getIdentity(phone):
        path = StorageTools.getStorageForPhone(phone)
        out = None
        idPath = os.path.join(path, "id")
        if os.path.isfile(idPath):
            with open(idPath, 'rb') as idFile:
                out = idFile.read()
        return out

    @staticmethod
    def writeNonce(phone, nonce):
        path = StorageTools.getStorageForPhone(phone)
        with open(os.path.join(path, "nonce"), 'wb') as idFile:
            idFile.write(nonce.encode("latin-1") if sys.version_info >= (3,0) else nonce)

    @staticmethod
    def getNonce(phone):
        path = StorageTools.getStorageForPhone(phone)
        out = None
        noncePath = os.path.join(path, "nonce")
        if os.path.isfile(noncePath):
            with open(noncePath, 'rb') as idFile:
                out = idFile.read()
        return out

class TimeTools:
    @staticmethod
    def parseIso(iso):
        d=datetime.datetime(*map(int, re.split('[^\d]', iso)[:-1]))
        return d

    @staticmethod
    def utcToLocal(dt):
        utc = tz.gettz('UTC')
        local = tz.tzlocal()
        dtUtc =  dt.replace(tzinfo=utc)

        return dtUtc.astimezone(local)

    @staticmethod
    def utcTimestamp():
        #utc = tz.gettz('UTC')
        utcNow = datetime.datetime.utcnow()
        return TimeTools.datetimeToTimestamp(utcNow)

    @staticmethod
    def datetimeToTimestamp(dt):
        return time.mktime(dt.timetuple())

class ImageTools:
    @staticmethod
    def scaleImage(infile, outfile, imageFormat, width, height):
        with PILOptionalModule() as imp:
            Image = imp("Image")
            im = Image.open(infile)
            #Convert P mode images
            if im.mode != "RGB":
                im = im.convert("RGB")
            im.thumbnail((width, height))
            im.save(outfile, imageFormat)
            return True
        return False

    @staticmethod
    def getImageDimensions(imageFile):
        with PILOptionalModule() as imp:
            Image = imp("Image")
            im = Image.open(imageFile)
            return im.size

    @staticmethod
    def generatePreviewFromImage(image):
        fd, path = tempfile.mkstemp()

        preview = None
        if ImageTools.scaleImage(image, path, "JPEG", YowConstants.PREVIEW_WIDTH, YowConstants.PREVIEW_HEIGHT):
            fileObj = os.fdopen(fd, "rb+")
            fileObj.seek(0)
            preview = fileObj.read()
            fileObj.close()
        os.remove(path)
        return preview

class MimeTools:
    MIME_FILE = os.path.join(os.path.dirname(__file__), 'mime.types')
    mimetypes.init() # Load default mime.types
    mimetypes.init([MIME_FILE]) # Append whatsapp mime.types
    decode_hex = codecs.getdecoder("hex_codec")

    @staticmethod
    def getMIME(filepath):
        mimeType = mimetypes.guess_type(filepath)[0]
        if mimeType is None:
            raise Exception("Unsupported/unrecognized file type for: "+filepath);
        return mimeType

class VideoTools:
    @staticmethod
    def getVideoProperties(videoFile):
        with FFVideoOptionalModule() as imp:
            VideoStream = imp("VideoStream")
            s = VideoStream(videoFile)
            return s.width, s.height, s.bitrate, s.duration #, s.codec_name

    @staticmethod
    def generatePreviewFromVideo(videoFile):
        with FFVideoOptionalModule() as imp:
            VideoStream = imp("VideoStream")
            fd, path = tempfile.mkstemp('.jpg')
            stream = VideoStream(videoFile)
            stream.get_frame_at_sec(0).image().save(path)
            preview = ImageTools.generatePreviewFromImage(path)
            os.remove(path)
            return preview



from yowsup.layers.auth.protocolentities.response import ResponseProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.structs.protocolentity import ProtocolEntityTest
import unittest

class ResponseProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = ResponseProtocolEntity
        attribs = {
            "xmlns": "urn:ietf:params:xml:ns:xmpp-sasl"
        }
        self.node = ProtocolTreeNode("response", attribs, None, "dummydata")


from yowsup.layers.protocol_iq.protocolentities import IqProtocolEntity
class PictureIqProtocolEntity(IqProtocolEntity):
    '''
    When receiving a profile picture:
    <iq type="result" from="{{jid}}" id="{{id}}">
        <picture type="image" id="{{another_id}}">
        {{Binary bytes of the picture.}}
        </picture>
    </iq>
    '''
    XMLNS = "w:profile:picture"

    def __init__(self, jid, _id = None, type = "get"):
        super(PictureIqProtocolEntity, self).__init__(self.__class__.XMLNS, _id = _id, _type=type, to = jid)


from yowsup.layers.interface                           import YowInterfaceLayer, ProtocolEntityCallback
from yowsup.layers.protocol_messages.protocolentities  import TextMessageProtocolEntity
from yowsup.common.tools import Jid
import threading
import logging
logger = logging.getLogger(__name__)

class SendLayer(YowInterfaceLayer):

    #This message is going to be replaced by the @param message in YowsupSendStack construction
    #i.e. list of (jid, message) tuples
    PROP_MESSAGES = "org.openwhatsapp.yowsup.prop.sendclient.queue"
    
    
    def __init__(self):
        super(SendLayer, self).__init__()
        self.ackQueue = []
        self.lock = threading.Condition()

    #call back function when there is a successful connection to whatsapp server
    @ProtocolEntityCallback("success")
    def onSuccess(self, successProtocolEntity):
        self.lock.acquire()
        for target in self.getProp(self.__class__.PROP_MESSAGES, []):
            #getProp() is trying to retreive the list of (jid, message) tuples, if none exist, use the default []
            phone, message = target
            messageEntity = TextMessageProtocolEntity(message, to = Jid.normalize(phone))
            #append the id of message to ackQueue list
            #which the id of message will be deleted when ack is received.
            self.ackQueue.append(messageEntity.getId())
            self.toLower(messageEntity)
        self.lock.release()

    #after receiving the message from the target number, target number will send a ack to sender(us)
    @ProtocolEntityCallback("ack")
    def onAck(self, entity):
        self.lock.acquire()
        #if the id match the id in ackQueue, then pop the id of the message out
        if entity.getId() in self.ackQueue:
            self.ackQueue.pop(self.ackQueue.index(entity.getId()))
            
        if not len(self.ackQueue):
            self.lock.release()
            logger.info("Message sent")
            raise KeyboardInterrupt()

        self.lock.release()



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from yowsup.layers.protocol_notifications.protocolentities import NotificationProtocolEntity

class GroupsNotificationProtocolEntity(NotificationProtocolEntity):
    '''

    <notification notify="WhatsApp" id="{{id}}" t="1420402514" participant="{{participant_jiid}}" from="{{group_jid}}" type="w:gp2">
    </notification>

    '''

    def __init__(self, _id,  _from, timestamp, notify, participant, offline):
        super(GroupsNotificationProtocolEntity, self).__init__("w:gp2", _id, _from, timestamp, notify, offline)
        self.setParticipant(participant)
        self.setGroupId(_from)

    def setParticipant(self, participant):
        self._participant = participant

    def getParticipant(self, full = True):
        return self._participant if full else self._participant.split('@')[0]

    def getGroupId(self):
        return self.groupId

    def setGroupId(self, groupId):
        self.groupId = groupId


    def __str__(self):
        out = super(GroupsNotificationProtocolEntity, self).__str__()
        out += "Participant: %s\n" % self.getParticipant()
        return out

    def toProtocolTreeNode(self):
        node = super(GroupsNotificationProtocolEntity, self).toProtocolTreeNode()
        node.setAttribute("participant", self.getParticipant())
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(GroupsNotificationProtocolEntity, GroupsNotificationProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = GroupsNotificationProtocolEntity
        entity.setParticipant(node.getAttributeValue("participant"))
        entity.setGroupId(node.getAttributeValue("from"))
        return entity



from .chatstate import ChatstateProtocolEntity
class OutgoingChatstateProtocolEntity(ChatstateProtocolEntity):
    '''
    INCOMING

    <chatstate from="xxxxxxxxxxx@s.whatsapp.net">
    <{{composing|paused}}></{{composing|paused}}>
    </chatstate>

    OUTGOING

    <chatstate to="xxxxxxxxxxx@s.whatsapp.net">
    <{{composing|paused}}></{{composing|paused}}>
    </chatstate>
    '''

    def __init__(self, _state, _to):
        super(OutgoingChatstateProtocolEntity, self).__init__(_state)
        self.setOutgoingData(_to)

    def setOutgoingData(self, _to):
        self._to = _to
    
    def toProtocolTreeNode(self):
        node = super(OutgoingChatstateProtocolEntity, self).toProtocolTreeNode()
        node.setAttribute("to", self._to)
        return node

    def __str__(self):
        out  = super(OutgoingChatstateProtocolEntity, self).__str__()
        out += "To: %s\n" % self._to
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = ChatstateProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = OutgoingChatstateProtocolEntity
        entity.setOutgoingData(
            node.getAttributeValue("to"),
        )
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class ResponseProtocolEntity(ProtocolEntity):
    def __init__(self, data, xmlns = "urn:ietf:params:xml:ns:xmpp-sasl"):
        super(ResponseProtocolEntity, self).__init__("response")
        self.xmlns = xmlns
        self.data = data
    
    def toProtocolTreeNode(self):
        return self._createProtocolTreeNode({"xmlns": self.xmlns}, children = None, data = self.data)

    @staticmethod
    def fromProtocolTreeNode(node):
        return ResponseProtocolEntity(node.getData(), node.getAttributeValue("xmlns"))


from yowsup.layers import YowParallelLayer
import asyncore, time, logging, random
from yowsup.layers import YowLayer
from yowsup.layers.auth                        import YowCryptLayer, YowAuthenticationProtocolLayer
from yowsup.layers.coder                       import YowCoderLayer
from yowsup.layers.logger                      import YowLoggerLayer
from yowsup.layers.network                     import YowNetworkLayer
from yowsup.layers.protocol_messages           import YowMessagesProtocolLayer
from yowsup.layers.stanzaregulator             import YowStanzaRegulator
from yowsup.layers.protocol_media              import YowMediaProtocolLayer
from yowsup.layers.protocol_acks               import YowAckProtocolLayer
from yowsup.layers.protocol_receipts           import YowReceiptProtocolLayer
from yowsup.layers.protocol_groups             import YowGroupsProtocolLayer
from yowsup.layers.protocol_presence           import YowPresenceProtocolLayer
from yowsup.layers.protocol_ib                 import YowIbProtocolLayer
from yowsup.layers.protocol_notifications      import YowNotificationsProtocolLayer
from yowsup.layers.protocol_iq                 import YowIqProtocolLayer
from yowsup.layers.protocol_contacts           import YowContactsIqProtocolLayer
from yowsup.layers.protocol_chatstate          import YowChatstateProtocolLayer
from yowsup.layers.protocol_privacy            import YowPrivacyProtocolLayer
from yowsup.layers.protocol_profiles           import YowProfilesProtocolLayer
from yowsup.layers.protocol_calls import YowCallsProtocolLayer
from yowsup import env
from yowsup.common.constants import YowConstants
import inspect
try:
    import Queue
except ImportError:
    import queue as Queue
logger = logging.getLogger(__name__)


YOWSUP_PROTOCOL_LAYERS_BASIC = (
    YowAuthenticationProtocolLayer, YowMessagesProtocolLayer,
    YowReceiptProtocolLayer, YowAckProtocolLayer, YowPresenceProtocolLayer,
    YowIbProtocolLayer, YowIqProtocolLayer, YowNotificationsProtocolLayer,
    YowContactsIqProtocolLayer, YowChatstateProtocolLayer, YowCallsProtocolLayer

)

class YowStackBuilder(object):

    def __init__(self):
        self.layers = ()
        self._props = {}

    def setProp(self, key, value):
        self._props[key] = value

    def pushDefaultLayers(self, axolotl = False):
        defaultLayers = YowStackBuilder.getDefaultLayers(axolotl)
        self.layers += defaultLayers
        return self

    def push(self, yowLayer):
        self.layers += (yowLayer,)
        return self

    def pop(self):
        self.layers = self.layers[:-1]
        return self

    def build(self):
        return YowStack(self.layers, reversed = False, props = self._props)

    @staticmethod
    def getDefaultLayers(axolotl = False, groups = True, media = True, privacy = True, profiles = True):
        coreLayers = YowStackBuilder.getCoreLayers()
        protocolLayers = YowStackBuilder.getProtocolLayers(groups = groups, media=media, privacy=privacy, profiles=profiles)

        allLayers = coreLayers
        if axolotl:
            from yowsup.layers.axolotl import YowAxolotlLayer
            allLayers += (YowAxolotlLayer,)

        allLayers += (YowParallelLayer(protocolLayers),)

        return allLayers

    @staticmethod
    def getDefaultStack(layer = None, axolotl = False, groups = True, media = True, privacy = True, profiles = True):
        """
        :param layer: An optional layer to put on top of default stack
        :param axolotl: E2E encryption enabled/ disabled
        :return: YowStack
        """

        allLayers = YowStackBuilder.getDefaultLayers(axolotl, groups = groups, media=media,privacy=privacy, profiles=profiles)
        if layer:
            allLayers = allLayers + (layer,)


        return YowStack(allLayers, reversed = False)

    @staticmethod
    def getCoreLayers():
        return (
            YowLoggerLayer,
            YowCoderLayer,
            YowCryptLayer,
            YowStanzaRegulator,
            YowNetworkLayer
        )[::-1]

    @staticmethod
    def getProtocolLayers(groups = True, media = True, privacy = True, profiles = True):
        layers = YOWSUP_PROTOCOL_LAYERS_BASIC
        if groups:
            layers += (YowGroupsProtocolLayer,)

        if media:
            layers += (YowMediaProtocolLayer, )

        if privacy:
            layers += (YowPrivacyProtocolLayer, )

        if profiles:
            layers += (YowProfilesProtocolLayer, )

        return layers

class YowStack(object):
    __stack = []
    __stackInstances = []
    __detachedQueue = Queue.Queue()
    def __init__(self, stackClassesArr = None, reversed = True, props = None):
        stackClassesArr = stackClassesArr or ()
        self.__stack = stackClassesArr[::-1] if reversed else stackClassesArr
        self.__stackInstances = []
        self._props = props or {}

        self.setProp(YowNetworkLayer.PROP_ENDPOINT, YowConstants.ENDPOINTS[random.randint(0,len(YowConstants.ENDPOINTS)-1)])
        self.setProp(YowCoderLayer.PROP_DOMAIN, YowConstants.DOMAIN)
        self.setProp(YowCoderLayer.PROP_RESOURCE, env.CURRENT_ENV.getResource())
        self._construct()


    def getLayerInterface(self, YowLayerClass):
        for inst in self.__stackInstances:
            if inst.__class__ == YowLayerClass:
                return inst.getLayerInterface()
            elif inst.__class__ == YowParallelLayer:
                res = inst.getLayerInterface(YowLayerClass)
                if res:
                    return res


    def send(self, data):
        self.__stackInstances[-1].send(data)

    def receive(self, data):
        self.__stackInstances[0].receive(data)

    def setCredentials(self, credentials):
        self.getLayerInterface(YowAuthenticationProtocolLayer).setCredentials(*credentials)

    def addLayer(self, layerClass):
        self.__stack.push(layerClass)

    def addPostConstructLayer(self, layer):
        self.__stackInstances[-1].setLayers(layer, self.__stackInstances[-2])
        layer.setLayers(None, self.__stackInstances[-1])
        self.__stackInstances.append(layer)

    def setProp(self, key, value):
        self._props[key] = value

    def getProp(self, key, default = None):
        return self._props[key] if key in self._props else default

    def emitEvent(self, yowLayerEvent):
        if not self.__stackInstances[0].onEvent(yowLayerEvent):
            self.__stackInstances[0].emitEvent(yowLayerEvent)

    def broadcastEvent(self, yowLayerEvent):
        if not self.__stackInstances[-1].onEvent(yowLayerEvent):
            self.__stackInstances[-1].broadcastEvent(yowLayerEvent)

    def execDetached(self, fn):
        self.__class__.__detachedQueue.put(fn)

    def loop(self, *args, **kwargs):
        if "discrete" in kwargs:
            discreteVal = kwargs["discrete"]
            del kwargs["discrete"]
            while True:
                asyncore.loop(*args, **kwargs)
                time.sleep(discreteVal)
                try:
                    callback = self.__class__.__detachedQueue.get(False) #doesn't block
                    callback()
                except Queue.Empty:
                    pass
        else:
            asyncore.loop(*args, **kwargs)

    def _construct(self):
        logger.debug("Initializing stack")
        for s in self.__stack:
            if type(s) is tuple:
                logger.warn("Implicit declaration of parallel layers in a tuple is deprecated, pass a YowParallelLayer instead")
                inst = YowParallelLayer(s)
            else:
                if inspect.isclass(s):
                    if issubclass(s, YowLayer):
                        inst = s()
                    else:
                        raise ValueError("Stack must contain only subclasses of YowLayer")
                elif issubclass(s.__class__, YowLayer):
                        inst = s
                else:
                    raise ValueError("Stack must contain only subclasses of YowLayer")
                #inst = s()
            logger.debug("Constructed %s" % inst)
            inst.setStack(self)
            self.__stackInstances.append(inst)

        for i in range(0, len(self.__stackInstances)):
            upperLayer = self.__stackInstances[i + 1] if (i + 1) < len(self.__stackInstances) else None
            lowerLayer = self.__stackInstances[i - 1] if i > 0 else None
            self.__stackInstances[i].setLayers(upperLayer, lowerLayer)

    def getLayer(self, layerIndex):
        return self.__stackInstances[layerIndex]



from yowsup.layers.protocol_notifications.protocolentities.notification_status import StatusNotificationProtocolEntity
from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_notifications.protocolentities.test_notification import NotificationProtocolEntityTest

class StatusNotificationProtocolEntityTest(NotificationProtocolEntityTest):
    def setUp(self):
        super(StatusNotificationProtocolEntityTest, self).setUp()
        self.ProtocolEntity = StatusNotificationProtocolEntity
        setNode = ProtocolTreeNode("set", {}, [], "status_data")
        self.node.addChild(setNode)



from .notification                  import NotificationProtocolEntity
from .notification_picture          import PictureNotificationProtocolEntity
from .notification_picture_set      import SetPictureNotificationProtocolEntity
from .notification_picture_delete      import DeletePictureNotificationProtocolEntity
from .notification_status           import StatusNotificationProtocolEntity



'''
Copyright (c) <2012> Tarek Galal <tare2.galal@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR
A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

import os, base64

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

class HttpProxy:

    def __init__(self, address, username = None, password = None):
        self.address = address
        self.username = username
        self.password = password

    def __repr__(self):
        return repr(self.address)

    def handler(self):
        return HttpProxyHandler(self)

    @staticmethod
    def getFromEnviron():
        url = None
        for key in ('http_proxy', 'https_proxy'):
            url = os.environ.get(key)
            if url: break
        if not url:
            return None
        dat = urlparse(url)
        port = 80 if dat.scheme == 'http' else 443
        if dat.port != None: port = int(dat.port)
        host = dat.hostname
        return HttpProxy((host, port), dat.username, dat.password)

class HttpProxyHandler:

    def __init__(self, proxy):
        self.state = 'init'
        self.proxy = proxy

    def onConnect(self):
        pass

    def connect(self, socket, pair):
        proxy = self.proxy
        authHeader = None
        if proxy.username and proxy.password:
            key = bytes(proxy.username, 'ascii') + b':' + bytes(proxy.password, 'ascii') if (bytes != str) else bytes(proxy.username) + b':' + proxy.password
            auth = base64.b64encode(key)
            authHeader = b'Proxy-Authorization: Basic ' + auth + b'\r\n'
        data = bytearray('CONNECT %s:%d HTTP/1.1\r\nHost: %s:%d\r\n' % (2 * pair), 'ascii')
        if authHeader:
            data += authHeader
        data += b'\r\n'
        self.state = 'connect'
        self.data = data
        socket.connect(proxy.address)

    def send(self, socket):
        if self.state == 'connect':
            socket.send(self.data)
            self.state = 'sent'

    def recv(self, socket, size):
        if self.state == 'sent':
            data = socket.recv(size)
            data = data.decode('ascii')
            status = data.split(' ', 2)
            if status[1] != '200':
                raise Exception('%s' % (data[:data.index('\r\n')]))
            self.state = 'end'
            self.onConnect()
            return data


class Group(object):
    def __init__(self, groupId, creatorJid, subject, subjectOwnerJid, subjectTime, creationTime, participants=None):
        self._groupId           = groupId
        self._creatorJid        = creatorJid
        self._subject           = subject
        self._subjectOwnerJid   = subjectOwnerJid
        self._subjectTime       = int(subjectTime)
        self._creationTime      = int(creationTime)
        self._participants      = participants or {}

    def getId(self):
        return self._groupId

    def getCreator(self):
        return self._creatorJid

    def getOwner(self):
        return self.getCreator()

    def getSubject(self):
        return self._subject

    def getSubjectOwner(self):
        return self._subjectOwnerJid

    def getSubjectTime(self):
        return self._subjectTime

    def getCreationTime(self):
        return self._creationTime

    def __str__(self):
        return "ID: %s, Subject: %s, Creation: %s, Creator: %s, Subject Owner: %s, Subject Time: %s\nParticipants: %s" %\
                (self.getId(), self.getSubject(), self.getCreationTime(), self.getCreator(),  self.getSubjectOwner(), self.getSubjectTime(), ", ".join(self._participants.keys()))

    def getParticipants(self):
        return self._participants



from .message_text import TextMessageProtocolEntity
from yowsup.structs import ProtocolTreeNode
import time
class BroadcastTextMessage(TextMessageProtocolEntity):
    def __init__(self, jids, body):
        broadcastTime = int(time.time() * 1000)
        super(BroadcastTextMessage, self).__init__(body, to = "%s@broadcast" % broadcastTime)
        self.setBroadcastProps(jids)

    def setBroadcastProps(self, jids):
        assert type(jids) is list, "jids must be a list, got %s instead." % type(jids)
        self.jids = jids

    def toProtocolTreeNode(self):
        node = super(BroadcastTextMessage, self).toProtocolTreeNode()
        toNodes = [ProtocolTreeNode("to", {"jid": jid}) for jid in self.jids]
        broadcastNode = ProtocolTreeNode("broadcast", children = toNodes)
        node.addChild(broadcastNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = TextMessageProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = BroadcastTextMessage
        jids = [toNode.getAttributeValue("jid") for toNode in node.getChild("broadcast").getAllChildren()]
        entity.setBroadcastProps(jids)
        return entity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
class SuccessProtocolEntity(ProtocolEntity):
    def __init__(self, status, kind, creation, expiration, props, t, nonce = None):
        super(SuccessProtocolEntity, self).__init__("success")
        self.status = status
        self.kind = kind
        self.creation = int(creation)
        self.expiration = int(expiration)
        self.props = props
        self.nonce = nonce
        self.t = int(t) ##whatever that is !

    def __str__(self):
        out  = "Account:\n"
        out += "Status: %s\n" % self.status
        out += "Kind: %s\n" % self.kind
        out += "Creation: %s\n" % self.creation
        out += "Expiration: %s\n" % self.expiration
        out += "Props: %s\n" % self.props
        out += "t: %s\n" % self.t
        return out

    def toProtocolTreeNode(self):
        attributes = {
            "status"      :    self.status,
            "kind"        :    self.kind,
            "creation"    :    str(self.creation),
            "expiration"  :    str(self.expiration),
            "props"       :    self.props,
            "t"           :    str(self.t)
        }
        return self._createProtocolTreeNode(attributes, children = None, data = self.nonce)

    @staticmethod
    def fromProtocolTreeNode(node):
        return SuccessProtocolEntity(
            node.getAttributeValue("status"),
            node.getAttributeValue("kind"),
            node.getAttributeValue("creation"),
            node.getAttributeValue("expiration"),
            node.getAttributeValue("props"),
            node.getAttributeValue("t"),
            node.getData()
            )


from .auth import AuthProtocolEntity
from .challenge import ChallengeProtocolEntity
from .response import ResponseProtocolEntity
from .stream_features import StreamFeaturesProtocolEntity
from .success import SuccessProtocolEntity
from .failure import FailureProtocolEntity
from .stream_error_conflict import StreamErrorConflictProtocolEntity
from .stream_error_ack import StreamErrorAckProtocolEntity



import abc
class YowsupEnv(object):
    __metaclass__ = abc.ABCMeta

    _USERAGENT_STRING = "WhatsApp/{WHATSAPP_VERSION} {OS_NAME}/{OS_VERSION} Device/{DEVICE_NAME}"

    @abc.abstractmethod
    def getToken(self, phoneNumber):
        pass

    @abc.abstractmethod
    def getVersion(self):
        pass

    @abc.abstractmethod
    def getOSVersion(self):
        pass

    @abc.abstractmethod
    def getOSName(self):
        pass

    @abc.abstractmethod
    def getDeviceName(self):
        pass

    @abc.abstractmethod
    def isAxolotlEnabled(self):
        pass

    def getResource(self):
        return self.getOSName() + "-" + self.getVersion()

    def getUserAgent(self):
        return self.__class__._USERAGENT_STRING.format(
            WHATSAPP_VERSION = self.getVersion(),
            OS_NAME = self.getOSName(),
            OS_VERSION = self.getOSVersion(),
            DEVICE_NAME = self.getDeviceName()
        )




from yowsup.common.http.warequest import WARequest
from yowsup.common.http.waresponseparser import JSONResponseParser
from yowsup.env import CURRENT_ENV

class WAExistsRequest(WARequest):
    
    def __init__(self,cc, p_in, idx):
        super(WAExistsRequest,self).__init__()

        self.addParam("cc", cc)
        self.addParam("in", p_in)
        self.addParam("id", idx)
        self.addParam("lg", "en")
        self.addParam("lc", "GB")
        self.addParam("token", CURRENT_ENV.getToken(p_in))

        self.url = "v.whatsapp.net/v2/exist"

        self.pvars = ["status", "reason", "sms_length", "voice_length", "result","param", "pw", "login", "type", "expiration", "kind",
                    "price", "cost", "currency", "price_expiration"
                    ]

        self.setParser(JSONResponseParser())

    def send(self, parser = None):
        res = super(WAExistsRequest, self).send(parser)
        return res


from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_groups import GroupsIqProtocolEntity

class ParticipantsGroupsIqProtocolEntity(GroupsIqProtocolEntity):
    '''
    <iq type="get" id="{{id}}" xmlns="w:g2", to={{group_jid}}">
        <list></list>
    </iq>
    '''

    modes=["add","promote","remove","demote"]
    def __init__(self, jid, participantList, _mode, _id = None):
        super(ParticipantsGroupsIqProtocolEntity, self).__init__(to = jid, _id = _id, _type = "set")
        self.setProps(group_jid = jid, participantList = participantList, mode = _mode)

    def setProps(self, group_jid, participantList, mode):
        assert type(participantList) is list, "Must be a list of jids, got %s instead." % type(participantList)
        assert mode in self.modes, "Mode should be in: '" + "', '".join(self.modes) + "' but is '" + mode + "'"
        self.group_jid = group_jid
        self.participantList = participantList
        self.mode = mode

    def toProtocolTreeNode(self):
        node = super(ParticipantsGroupsIqProtocolEntity, self).toProtocolTreeNode()
        participantNodes = [
            ProtocolTreeNode("participant", {
                "jid":       participant
            })
            for participant in self.participantList
        ]
        node.addChild(ProtocolTreeNode(self.mode,{}, participantNodes))
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(ParticipantsGroupsIqProtocolEntity, ParticipantsGroupsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = ParticipantsGroupsIqProtocolEntity
        return entity



from yowsup.structs import ProtocolTreeNode
from .notification_contact import ContactNotificationProtocolEntity
class ContactsSyncNotificationProtocolEntity(ContactNotificationProtocolEntity):
    '''
    <notification from="4917667738517@s.whatsapp.net" t="1437251557" offline="0" type="contacts" id="4174521704">
        <sync after="1437251557"></sync>
    </notification>
    '''

    def __init__(self, _id,  _from, timestamp, notify, offline, after):
        super(ContactsSyncNotificationProtocolEntity, self).__init__(_id, _from, timestamp, notify, offline)
        self.setData(after)

    def setData(self, after):
        self.after = int(after)

    def toProtocolTreeNode(self):
        node = super(ContactsSyncNotificationProtocolEntity, self).toProtocolTreeNode()
        syncNode = ProtocolTreeNode("sync", {"after": str(self.after)}, None, None)
        node.addChild(syncNode)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = ContactNotificationProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = ContactsSyncNotificationProtocolEntity
        syncNode = node.getChild("sync")
        entity.setData(syncNode.getAttributeValue("after"))
        return entity


from yowsup.structs import ProtocolTreeNode
from yowsup.layers.protocol_iq.protocolentities import ResultIqProtocolEntity
class ListParticipantsResultIqProtocolEntity(ResultIqProtocolEntity):
    '''
    <iq type="result" from="{{GROUP_ID}}" id="{{IQ_ID}}">
        <participant jid="{{PARTICIPANT_JID}}">
        </participant>
    </iq>
    '''

    def __init__(self, _from, participantList):
        super(ListParticipantsResultIqProtocolEntity, self).__init__(_from = _from)
        self.setParticipants(participantList)

    def __str__(self):
        out = super(ListParticipantsResultIqProtocolEntity, self).__str__()
        out += "Participants: %s\n" % " ".join(self.participantList)
        return out

    def getParticipants(self):
        return self.participantList

    def setParticipants(self, participants):
        self.participantList = participants

    def toProtocolTreeNode(self):
        node = super(ListParticipantsResultIqProtocolEntity, self).toProtocolTreeNode()

        participantNodes = [
            ProtocolTreeNode("participant", {
                "jid":       participant
            })
            for participant in self.participantList
        ]

        node.addChildren(participantNodes)
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = super(ListParticipantsResultIqProtocolEntity, ListParticipantsResultIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = ListParticipantsResultIqProtocolEntity
        entity.setParticipants([ pNode.getAttributeValue("jid") for pNode in node.getAllChildren() ])
        return entity



from .protocoltreenode import ProtocolTreeNode
import unittest, time
class ProtocolEntity(object):
    __ID_GEN = 0

    def __init__(self, tag):
        self.tag = tag

    def getTag(self):
        return self.tag

    def isType(self,  typ):
        return self.tag == typ

    def _createProtocolTreeNode(self, attributes, children = None, data = None):
        return ProtocolTreeNode(self.getTag(), attributes, children, data)


    def _getCurrentTimestamp(self):
        return int(time.time())

    def _generateId(self, short = False):
        ProtocolEntity.__ID_GEN += 1
        return str(ProtocolEntity.__ID_GEN) if short else str(int(time.time())) + "-" + str(ProtocolEntity.__ID_GEN)


    def toProtocolTreeNode(self):
        pass

    @staticmethod
    def fromProtocolTreeNode(self, protocolTreeNode):
        pass


class ProtocolEntityTest(object):
    def setUp(self):
        self.ProtocolEntity = None
        self.node = None

    # def assertEqual(self, entity, node):
    #     raise AssertionError("Should never execute that")

    def test_generation(self):
        if self.ProtocolEntity is None:
            raise ValueError("Test case not setup!")
        entity = self.ProtocolEntity.fromProtocolTreeNode(self.node)
        try:
            self.assertEqual(entity.toProtocolTreeNode(), self.node)
        except:
            print(entity.toProtocolTreeNode())
            print("\nNOTEQ\n")
            print(self.node)
            raise




from .cli import Cli, clicmd
from yowsup.layers.interface import YowInterfaceLayer, ProtocolEntityCallback
from yowsup.layers.auth import YowAuthenticationProtocolLayer
from yowsup.layers import YowLayerEvent, EventCallback
from yowsup.layers.network import YowNetworkLayer
import sys
from yowsup.common import YowConstants
import datetime
import os
import logging
from yowsup.layers.protocol_receipts.protocolentities    import *
from yowsup.layers.protocol_groups.protocolentities      import *
from yowsup.layers.protocol_presence.protocolentities    import *
from yowsup.layers.protocol_messages.protocolentities    import *
from yowsup.layers.protocol_acks.protocolentities        import *
from yowsup.layers.protocol_ib.protocolentities          import *
from yowsup.layers.protocol_iq.protocolentities          import *
from yowsup.layers.protocol_contacts.protocolentities    import *
from yowsup.layers.protocol_chatstate.protocolentities   import *
from yowsup.layers.protocol_privacy.protocolentities     import *
from yowsup.layers.protocol_media.protocolentities       import *
from yowsup.layers.protocol_media.mediauploader import MediaUploader
from yowsup.layers.protocol_profiles.protocolentities    import *
from yowsup.common.tools import Jid
from yowsup.common.optionalmodules import PILOptionalModule, AxolotlOptionalModule

logger = logging.getLogger(__name__)

class YowsupCliLayer(Cli, YowInterfaceLayer):
    PROP_RECEIPT_AUTO       = "org.openwhatsapp.yowsup.prop.cli.autoreceipt"
    PROP_RECEIPT_KEEPALIVE  = "org.openwhatsapp.yowsup.prop.cli.keepalive"
    PROP_CONTACT_JID        = "org.openwhatsapp.yowsup.prop.cli.contact.jid"
    EVENT_LOGIN             = "org.openwhatsapp.yowsup.event.cli.login"
    EVENT_START             = "org.openwhatsapp.yowsup.event.cli.start"
    EVENT_SENDANDEXIT       = "org.openwhatsapp.yowsup.event.cli.sendandexit"

    MESSAGE_FORMAT          = "[{FROM}({TIME})]:[{MESSAGE_ID}]\t {MESSAGE}"

    FAIL_OPT_PILLOW         = "No PIL library installed, try install pillow"
    FAIL_OPT_AXOLOTL        = "axolotl is not installed, try install python-axolotl"

    DISCONNECT_ACTION_PROMPT = 0
    DISCONNECT_ACTION_EXIT   = 1

    ACCOUNT_DEL_WARNINGS = 4

    def __init__(self):
        super(YowsupCliLayer, self).__init__()
        YowInterfaceLayer.__init__(self)
        self.accountDelWarnings = 0
        self.connected = False
        self.username = None
        self.sendReceipts = True
        self.sendRead = True
        self.disconnectAction = self.__class__.DISCONNECT_ACTION_PROMPT
        self.credentials = None

        #add aliases to make it user to use commands. for example you can then do:
        # /message send foobar "HI"
        # and then it will get automaticlaly mapped to foobar's jid
        self.jidAliases = {
            # "NAME": "PHONE@s.whatsapp.net"
        }

    def aliasToJid(self, calias):
        for alias, ajid in self.jidAliases.items():
            if calias.lower() == alias.lower():
                return Jid.normalize(ajid)

        return Jid.normalize(calias)

    def jidToAlias(self, jid):
        for alias, ajid in self.jidAliases.items():
            if ajid == jid:
                return alias
        return jid

    def setCredentials(self, username, password):
        self.getLayerInterface(YowAuthenticationProtocolLayer).setCredentials(username, password)


    @EventCallback(EVENT_START)
    def onStart(self, layerEvent):
        self.startInput()
        return True

    @EventCallback(EVENT_SENDANDEXIT)
    def onSendAndExit(self, layerEvent):
        credentials = layerEvent.getArg("credentials")
        target = layerEvent.getArg("target")
        message = layerEvent.getArg("message")
        self.sendMessageAndDisconnect(credentials, target, message)
        return True

    @EventCallback(YowNetworkLayer.EVENT_STATE_DISCONNECTED)
    def onStateDisconnected(self,layerEvent):
        self.output("Disconnected: %s" % layerEvent.getArg("reason"))
        if self.disconnectAction == self.__class__.DISCONNECT_ACTION_PROMPT:
           self.connected = False
           self.notifyInputThread()
        else:
           os._exit(os.EX_OK)

    def assertConnected(self):
        if self.connected:
            return True
        else:
            self.output("Not connected", tag = "Error", prompt = False)
            return False

    #### batch cmds #####
    def sendMessageAndDisconnect(self, credentials, jid, message):
        self.disconnectAction = self.__class__.DISCONNECT_ACTION_EXIT
        self.queueCmd("/login %s %s" % credentials)
        self.queueCmd("/message send %s \"%s\" wait" % (jid, message))
        self.queueCmd("/disconnect")
        self.startInput()


    ########## PRESENCE ###############
    @clicmd("Set presence name")
    def presence_name(self, name):
        if self.assertConnected():
            entity = PresenceProtocolEntity(name = name)
            self.toLower(entity)

    @clicmd("Set presence as available")
    def presence_available(self):
        if self.assertConnected():
            entity = AvailablePresenceProtocolEntity()
            self.toLower(entity)

    @clicmd("Set presence as unavailable")
    def presence_unavailable(self):
        if self.assertConnected():
            entity = UnavailablePresenceProtocolEntity()
            self.toLower(entity)

    @clicmd("Unsubscribe from contact's presence updates")
    def presence_unsubscribe(self, contact):
        if self.assertConnected():
            entity = UnsubscribePresenceProtocolEntity(self.aliasToJid(contact))
            self.toLower(entity)

    @clicmd("Subscribe to contact's presence updates")
    def presence_subscribe(self, contact):
        if self.assertConnected():
            entity = SubscribePresenceProtocolEntity(self.aliasToJid(contact))
            self.toLower(entity)

    ########### END PRESENCE #############

    ########### ib #######################
    @clicmd("Send clean dirty")
    def ib_clean(self, dirtyType):
        if self.assertConnected():
            entity = CleanIqProtocolEntity("groups", YowConstants.DOMAIN)
            self.toLower(entity)

    @clicmd("Ping server")
    def ping(self):
        if self.assertConnected():
            entity = PingIqProtocolEntity(to = YowConstants.DOMAIN)
            self.toLower(entity)

    ######################################

    ####### contacts/ profiles ####################
    @clicmd("Set status text")
    def profile_setStatus(self, text):
        if self.assertConnected():

            def onSuccess(resultIqEntity, originalIqEntity):
                self.output("Status updated successfully")

            def onError(errorIqEntity, originalIqEntity):
                logger.error("Error updating status")

            entity = SetStatusIqProtocolEntity(text)
            self._sendIq(entity, onSuccess, onError)

    @clicmd("Get profile picture for contact")
    def contact_picture(self, jid):
        if self.assertConnected():
            entity = GetPictureIqProtocolEntity(self.aliasToJid(jid), preview=False)
            self._sendIq(entity, self.onGetContactPictureResult)

    @clicmd("Get profile picture preview for contact")
    def contact_picturePreview(self, jid):
        if self.assertConnected():
            entity = GetPictureIqProtocolEntity(self.aliasToJid(jid), preview=True)
            self._sendIq(entity, self.onGetContactPictureResult)

    @clicmd("Get lastseen for contact")
    def contact_lastseen(self, jid):
        if self.assertConnected():
            def onSuccess(resultIqEntity, originalIqEntity):
                self.output("%s lastseen %s seconds ago" % (resultIqEntity.getFrom(), resultIqEntity.getSeconds()))

            def onError(errorIqEntity, originalIqEntity):
                logger.error("Error getting lastseen information for %s" % originalIqEntity.getTo())

            entity = LastseenIqProtocolEntity(self.aliasToJid(jid))
            self._sendIq(entity, onSuccess, onError)

    @clicmd("Set profile picture")
    def profile_setPicture(self, path):
        if self.assertConnected():
            with PILOptionalModule(failMessage = "No PIL library installed, try install pillow") as imp:
                Image = imp("Image")
                def onSuccess(resultIqEntity, originalIqEntity):
                    self.output("Profile picture updated successfully")

                def onError(errorIqEntity, originalIqEntity):
                    logger.error("Error updating profile picture")

                #example by @aesedepece in https://github.com/tgalal/yowsup/pull/781
                #modified to support python3
                src = Image.open(path)
                pictureData = src.resize((640, 640)).tobytes("jpeg", "RGB")
                picturePreview = src.resize((96, 96)).tobytes("jpeg", "RGB")
                iq = SetPictureIqProtocolEntity(self.getOwnJid(), picturePreview, pictureData)
                self._sendIq(iq, onSuccess, onError)

    @clicmd("Get profile privacy")
    def profile_getPrivacy(self):
        if self.assertConnected():
            def onSuccess(resultIqEntity, originalIqEntity):
                self.output("Profile privacy is: %s" %(resultIqEntity))

            def onError(errorIqEntity, originalIqEntity):
                logger.error("Error getting profile privacy")

            iq = GetPrivacyIqProtocolEntity()
            self._sendIq(iq, onSuccess, onError)

    @clicmd("Profile privacy. value=all|contacts|none names=profile|status|last. Names are comma separated, defaults to all.")
    def profile_setPrivacy(self, value="all", names=None):
        if self.assertConnected():
            def onSuccess(resultIqEntity, originalIqEntity):
                self.output("Profile privacy set to: %s" %(resultIqEntity))

            def onError(errorIqEntity, originalIqEntity):
                logger.error("Error setting profile privacy")
            try:
                names = [name for name in names.split(',')] if names else None
                iq = SetPrivacyIqProtocolEntity(value, names)
                self._sendIq(iq, onSuccess, onError)
            except Exception as inst:
                self.output(inst.message)
                return self.print_usage()


    ########### groups

    @clicmd("List all groups you belong to", 5)
    def groups_list(self):
        if self.assertConnected():
            entity = ListGroupsIqProtocolEntity()
            self.toLower(entity)

    @clicmd("Leave a group you belong to", 4)
    def group_leave(self, group_jid):
        if self.assertConnected():
            entity = LeaveGroupsIqProtocolEntity([self.aliasToJid(group_jid)])
            self.toLower(entity)

    @clicmd("Create a new group with the specified subject and participants. Jids are a comma separated list but optional.", 3)
    def groups_create(self, subject, jids = None):
        if self.assertConnected():
            jids = [self.aliasToJid(jid) for jid in jids.split(',')] if jids else []
            entity = CreateGroupsIqProtocolEntity(subject, participants=jids)
            self.toLower(entity)

    @clicmd("Invite to group. Jids are a comma separated list")
    def group_invite(self, group_jid, jids):
        if self.assertConnected():
            jids = [self.aliasToJid(jid) for jid in jids.split(',')]
            entity = AddParticipantsIqProtocolEntity(self.aliasToJid(group_jid), jids)
            self.toLower(entity)

    @clicmd("Promote admin of a group. Jids are a comma separated list")
    def group_promote(self, group_jid, jids):
        if self.assertConnected():
            jids = [self.aliasToJid(jid) for jid in jids.split(',')]
            entity = PromoteParticipantsIqProtocolEntity(self.aliasToJid(group_jid), jids)
            self.toLower(entity)

    @clicmd("Remove admin of a group. Jids are a comma separated list")
    def group_demote(self, group_jid, jids):
        if self.assertConnected():
            jids = [self.aliasToJid(jid) for jid in jids.split(',')]
            entity = DemoteParticipantsIqProtocolEntity(self.aliasToJid(group_jid), jids)
            self.toLower(entity)

    @clicmd("Kick from group. Jids are a comma separated list")
    def group_kick(self, group_jid, jids):
        if self.assertConnected():
            jids = [self.aliasToJid(jid) for jid in jids.split(',')]
            entity = RemoveParticipantsIqProtocolEntity(self.aliasToJid(group_jid), jids)
            self.toLower(entity)

    @clicmd("Change group subject")
    def group_setSubject(self, group_jid, subject):
        if self.assertConnected():
            entity = SubjectGroupsIqProtocolEntity(self.aliasToJid(group_jid), subject)
            self.toLower(entity)

    @clicmd("Set group picture")
    def group_picture(self, group_jid, path):
        if self.assertConnected():
            with PILOptionalModule(failMessage = self.__class__.FAIL_OPT_PILLOW) as imp:
                Image = imp("Image")

                def onSuccess(resultIqEntity, originalIqEntity):
                    self.output("Group picture updated successfully")

                def onError(errorIqEntity, originalIqEntity):
                    logger.error("Error updating Group picture")

                #example by @aesedepece in https://github.com/tgalal/yowsup/pull/781
                #modified to support python3
                src = Image.open(path)
                pictureData = src.resize((640, 640)).tobytes("jpeg", "RGB")
                picturePreview = src.resize((96, 96)).tobytes("jpeg", "RGB")
                iq = SetPictureIqProtocolEntity(self.aliasToJid(group_jid), picturePreview, pictureData)
                self._sendIq(iq, onSuccess, onError)


    @clicmd("Get group info")
    def group_info(self, group_jid):
        if self.assertConnected():
            entity = InfoGroupsIqProtocolEntity(self.aliasToJid(group_jid))
            self.toLower(entity)

    @clicmd("Get shared keys")
    def keys_get(self, jids):
        with AxolotlOptionalModule(failMessage = self.__class__.FAIL_OPT_AXOLOTL) as importFn:
            importFn()
            from yowsup.layers.axolotl.protocolentities.iq_key_get import GetKeysIqProtocolEntity
            if self.assertConnected():
                jids = [self.aliasToJid(jid) for jid in jids.split(',')]
                entity = GetKeysIqProtocolEntity(jids)
                self.toLower(entity)

    @clicmd("Send prekeys")
    def keys_set(self):
        with AxolotlOptionalModule(failMessage = self.__class__.FAIL_OPT_AXOLOTL) as axoOptMod:
            from yowsup.layers.axolotl import YowAxolotlLayer
            if self.assertConnected():
                self.broadcastEvent(YowLayerEvent(YowAxolotlLayer.EVENT_PREKEYS_SET))

    @clicmd("Send init seq")
    def seq(self):
        priv = PrivacyListIqProtocolEntity()
        self.toLower(priv)
        push = PushIqProtocolEntity()
        self.toLower(push)
        props = PropsIqProtocolEntity()
        self.toLower(props)
        crypto = CryptoIqProtocolEntity()
        self.toLower(crypto)

    @clicmd("Delete your account")
    def account_delete(self):
        if self.assertConnected():
            if self.accountDelWarnings < self.__class__.ACCOUNT_DEL_WARNINGS:
                self.accountDelWarnings += 1
                remaining = self.__class__.ACCOUNT_DEL_WARNINGS - self.accountDelWarnings
                self.output("Repeat delete command another %s times to send the delete request" % remaining, tag="Account delete Warning !!", prompt = False)
            else:
                entity = UnregisterIqProtocolEntity()
                self.toLower(entity)

    @clicmd("Send message to a friend")
    def message_send(self, number, content):
        if self.assertConnected():
            outgoingMessage = TextMessageProtocolEntity(content.encode("utf-8") if sys.version_info >= (3,0) else content, to = self.aliasToJid(number))
            self.toLower(outgoingMessage)

    @clicmd("Broadcast message. numbers should comma separated phone numbers")
    def message_broadcast(self, numbers, content):
        if self.assertConnected():
            jids = [self.aliasToJid(number) for number in numbers.split(',')]
            outgoingMessage = BroadcastTextMessage(jids, content)
            self.toLower(outgoingMessage)

    #@clicmd("Send read receipt")
    def message_read(self, message_id):
        pass

    #@clicmd("Send delivered receipt")
    def message_delivered(self, message_id):
        pass

    @clicmd("Send a video with optional caption")
    def video_send(self, number, path, caption = None):
        self.media_send(number, path, RequestUploadIqProtocolEntity.MEDIA_TYPE_VIDEO)

    @clicmd("Send an image with optional caption")
    def image_send(self, number, path, caption = None):
        self.media_send(number, path, RequestUploadIqProtocolEntity.MEDIA_TYPE_IMAGE)

    @clicmd("Send audio file")
    def audio_send(self, number, path):
        self.media_send(number, path, RequestUploadIqProtocolEntity.MEDIA_TYPE_AUDIO)

    def media_send(self, number, path, mediaType, caption = None):
        if self.assertConnected():
            jid = self.aliasToJid(number)
            entity = RequestUploadIqProtocolEntity(mediaType, filePath=path)
            successFn = lambda successEntity, originalEntity: self.onRequestUploadResult(jid, mediaType, path, successEntity, originalEntity, caption)
            errorFn = lambda errorEntity, originalEntity: self.onRequestUploadError(jid, path, errorEntity, originalEntity)
            self._sendIq(entity, successFn, errorFn)

            self._sendIq(entity, successFn, errorFn)
    @clicmd("Send typing state")
    def state_typing(self, jid):
        if self.assertConnected():
            entity = OutgoingChatstateProtocolEntity(ChatstateProtocolEntity.STATE_TYPING, self.aliasToJid(jid))
            self.toLower(entity)

    @clicmd("Request contacts statuses")
    def statuses_get(self, contacts):
        if self.assertConnected():
            entity = GetStatusesIqProtocolEntity([self.aliasToJid(c) for c in contacts.split(',')])
            self.toLower(entity)

    @clicmd("Send paused state")
    def state_paused(self, jid):
        if self.assertConnected():
            entity = OutgoingChatstateProtocolEntity(ChatstateProtocolEntity.STATE_PAUSED, self.aliasToJid(jid))
            self.toLower(entity)

    @clicmd("Sync contacts, contacts should be comma separated phone numbers, with no spaces")
    def contacts_sync(self, contacts):
        if self.assertConnected():
            entity = GetSyncIqProtocolEntity(contacts.split(','))
            self.toLower(entity)

    @clicmd("Disconnect")
    def disconnect(self):
        if self.assertConnected():

            self.broadcastEvent(YowLayerEvent(YowNetworkLayer.EVENT_STATE_DISCONNECT))

    @clicmd("Quick login")
    def L(self):
        if self.connected:
            return self.output("Already connected, disconnect first")
        self.getLayerInterface(YowNetworkLayer).connect()
        return True

    @clicmd("Login to WhatsApp", 0)
    def login(self, username, b64password):
        self.setCredentials(username, b64password)
        return self.L()

    ######## receive #########

    @ProtocolEntityCallback("chatstate")
    def onChatstate(self, entity):
        print(entity)

    @ProtocolEntityCallback("iq")
    def onIq(self, entity):
        print(entity)

    @ProtocolEntityCallback("receipt")
    def onReceipt(self, entity):
        self.toLower(entity.ack())

    @ProtocolEntityCallback("ack")
    def onAck(self, entity):
        #formattedDate = datetime.datetime.fromtimestamp(self.sentCache[entity.getId()][0]).strftime('%d-%m-%Y %H:%M')
        #print("%s [%s]:%s"%(self.username, formattedDate, self.sentCache[entity.getId()][1]))
        if entity.getClass() == "message":
            self.output(entity.getId(), tag = "Sent")
            #self.notifyInputThread()

    @ProtocolEntityCallback("success")
    def onSuccess(self, entity):
        self.connected = True
        self.output("Logged in!", "Auth", prompt = False)
        self.notifyInputThread()

    @ProtocolEntityCallback("failure")
    def onFailure(self, entity):
        self.connected = False
        self.output("Login Failed, reason: %s" % entity.getReason(), prompt = False)

    @ProtocolEntityCallback("notification")
    def onNotification(self, notification):
        notificationData = notification.__str__()
        if notificationData:
            self.output(notificationData, tag = "Notification")
        else:
            self.output("From :%s, Type: %s" % (self.jidToAlias(notification.getFrom()), notification.getType()), tag = "Notification")
        if self.sendReceipts:
            self.toLower(notification.ack())

    @ProtocolEntityCallback("message")
    def onMessage(self, message):
        messageOut = ""
        if message.getType() == "text":
            #self.output(message.getBody(), tag = "%s [%s]"%(message.getFrom(), formattedDate))
            messageOut = self.getTextMessageBody(message)
        elif message.getType() == "media":
            messageOut = self.getMediaMessageBody(message)
        else:
            messageOut = "Unknown message type %s " % message.getType()
            print(messageOut.toProtocolTreeNode())


        formattedDate = datetime.datetime.fromtimestamp(message.getTimestamp()).strftime('%d-%m-%Y %H:%M')
        sender = message.getFrom() if not message.isGroupMessage() else "%s/%s" % (message.getParticipant(False), message.getFrom())
        output = self.__class__.MESSAGE_FORMAT.format(
            FROM = sender,
            TIME = formattedDate,
            MESSAGE = messageOut.encode('latin-1').decode() if sys.version_info >= (3, 0) else messageOut,
            MESSAGE_ID = message.getId()
            )

        self.output(output, tag = None, prompt = not self.sendReceipts)
        if self.sendReceipts:
            self.toLower(message.ack(self.sendRead))
            self.output("Sent delivered receipt"+" and Read" if self.sendRead else "", tag = "Message %s" % message.getId())


    def getTextMessageBody(self, message):
        return message.getBody()

    def getMediaMessageBody(self, message):
        if message.getMediaType() in ("image", "audio", "video"):
            return self.getDownloadableMediaMessageBody(message)
        else:
            return "[Media Type: %s]" % message.getMediaType()


    def getDownloadableMediaMessageBody(self, message):
         return "[Media Type:{media_type}, Size:{media_size}, URL:{media_url}]".format(
            media_type = message.getMediaType(),
            media_size = message.getMediaSize(),
            media_url = message.getMediaUrl()
            )

    def doSendMedia(self, mediaType, filePath, url, to, ip = None, caption = None):
        if mediaType == RequestUploadIqProtocolEntity.MEDIA_TYPE_IMAGE:
        	entity = ImageDownloadableMediaMessageProtocolEntity.fromFilePath(filePath, url, ip, to, caption = caption)
        elif mediaType == RequestUploadIqProtocolEntity.MEDIA_TYPE_AUDIO:
        	entity = AudioDownloadableMediaMessageProtocolEntity.fromFilePath(filePath, url, ip, to)
        elif mediaType == RequestUploadIqProtocolEntity.MEDIA_TYPE_VIDEO:
        	entity = VideoDownloadableMediaMessageProtocolEntity.fromFilePath(filePath, url, ip, to, caption = caption)
        self.toLower(entity)

    def __str__(self):
        return "CLI Interface Layer"

    ########### callbacks ############

    def onRequestUploadResult(self, jid, mediaType, filePath, resultRequestUploadIqProtocolEntity, requestUploadIqProtocolEntity, caption = None):

        if resultRequestUploadIqProtocolEntity.isDuplicate():
            self.doSendMedia(mediaType, filePath, resultRequestUploadIqProtocolEntity.getUrl(), jid,
                             resultRequestUploadIqProtocolEntity.getIp(), caption)
        else:
            successFn = lambda filePath, jid, url: self.doSendMedia(mediaType, filePath, url, jid, resultRequestUploadIqProtocolEntity.getIp(), caption)
            mediaUploader = MediaUploader(jid, self.getOwnJid(), filePath,
                                      resultRequestUploadIqProtocolEntity.getUrl(),
                                      resultRequestUploadIqProtocolEntity.getResumeOffset(),
                                      successFn, self.onUploadError, self.onUploadProgress, async=False)
            mediaUploader.start()

    def onRequestUploadError(self, jid, path, errorRequestUploadIqProtocolEntity, requestUploadIqProtocolEntity):
        logger.error("Request upload for file %s for %s failed" % (path, jid))

    def onUploadError(self, filePath, jid, url):
        logger.error("Upload file %s to %s for %s failed!" % (filePath, url, jid))

    def onUploadProgress(self, filePath, jid, url, progress):
        sys.stdout.write("%s => %s, %d%% \r" % (os.path.basename(filePath), jid, progress))
        sys.stdout.flush()

    def onGetContactPictureResult(self, resultGetPictureIqProtocolEntiy, getPictureIqProtocolEntity):
        # do here whatever you want
        # write to a file
        # or open
        # or do nothing
        # write to file example:
        #resultGetPictureIqProtocolEntiy.writeToFile("/tmp/yowpics/%s_%s.jpg" % (getPictureIqProtocolEntity.getTo(), "preview" if resultGetPictureIqProtocolEntiy.isPreview() else "full"))
        pass

    def __str__(self):
        return "CLI Interface Layer"

    @clicmd("Print this message")
    def help(self):
        self.print_usage()



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from yowsup.layers.protocol_receipts.protocolentities import OutgoingReceiptProtocolEntity
from yowsup.layers.axolotl.protocolentities.iq_keys_get_result import ResultGetKeysIqProtocolEntity
class RetryOutgoingReceiptProtocolEntity(OutgoingReceiptProtocolEntity):

    '''
    <receipt type="retry" to="xxxxxxxxxxx@s.whatsapp.net" id="1415389947-12" t="1432833777">
        <retry count="1" t="1432833266" id="1415389947-12" v="1">
        </retry>
        <registration>
            HEX:xxxxxxxxx
        </registration>
    </receipt>

    '''

    def __init__(self, _id, to, t, v = "1", count = "1",regData = ""):
        super(RetryOutgoingReceiptProtocolEntity, self).__init__(_id,to)
        self.setRetryData(t,v,count,regData)

    def setRetryData(self, t,v,count,regData):
        self.t = int(t)
        self.v = int(v)
        self.count = int(count)
        self.regData = regData

    def setRegData(self,regData):
        '''
        In axolotl layer:
        regData = self.store.getLocalRegistrationId()
        '''
        self.regData = ResultGetKeysIqProtocolEntity._intToBytes(regData)

    def toProtocolTreeNode(self):
        node = super(RetryOutgoingReceiptProtocolEntity, self).toProtocolTreeNode()
        node.setAttribute("type", "retry")
        retry = ProtocolTreeNode("retry", {"count": str(self.count),"t":str(self.t),"id":self.getId(),"v":str(self.v)})
        node.addChild(retry)
        registration = ProtocolTreeNode("registration",data=self.regData)
        node.addChild(registration)
        return node

    def __str__(self):
        out = super(RetryOutgoingReceiptProtocolEntity, self).__str__()
        return out

    @staticmethod
    def fromProtocolTreeNode(node):
        entity = OutgoingReceiptProtocolEntity.fromProtocolTreeNode(node)
        entity.__class__ = RetryOutgoingReceiptProtocolEntity
        retryNode = node.getChild("retry")
        entity.setRetryData(retryNode["t"], retryNode["v"], retryNode["count"], node.getChild("registration").data)


    @staticmethod
    def fromMesageNode(MessageNodeToBeRetried):
        return RetryOutgoingReceiptProtocolEntity(
            MessageNodeToBeRetried.getAttributeValue("id"),
            MessageNodeToBeRetried.getAttributeValue("from"),
            MessageNodeToBeRetried.getAttributeValue("t"),
            MessageNodeToBeRetried.getChild("enc").getAttributeValue("v")
        )


from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest
from yowsup.layers.protocol_iq.protocolentities import ErrorIqProtocolEntity
from yowsup.structs import ProtocolTreeNode

class ErrorIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(ErrorIqProtocolEntityTest, self).setUp()
        self.ProtocolEntity = ErrorIqProtocolEntity
        errorNode = ProtocolTreeNode("error", {"code": "123", "text": "abc"})
        self.node.addChild(errorNode)



from yowsup.layers.protocol_groups.protocolentities.iq_result_groups_list import ListGroupsResultIqProtocolEntity
from yowsup.structs.protocolentity import ProtocolEntityTest
from yowsup.layers.protocol_groups.structs import Group
import unittest
import time

entity = ListGroupsResultIqProtocolEntity(
    [
        Group("1234-456", "owner@s.whatsapp.net", "subject", "sOwnerJid@s.whatsapp.net", int(time.time()), int(time.time())),
        Group("4321-456", "owner@s.whatsapp.net", "subject", "sOwnerJid@s.whatsapp.net", int(time.time()), int(time.time()))
    ]
)

class ListGroupsResultIqProtocolEntityTest(ProtocolEntityTest, unittest.TestCase):
    def setUp(self):
        self.ProtocolEntity = ListGroupsResultIqProtocolEntity
        self.node = entity.toProtocolTreeNode()



from yowsup.common import YowConstants
from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from .iq_groups import GroupsIqProtocolEntity
class LeaveGroupsIqProtocolEntity(GroupsIqProtocolEntity):
    '''
    <iq id="{{id}}"" type="set" to="{{group_jid}}" xmlns="w:g2">
        <leave action="delete">
		    <group id="{{JID}}">
            <group id="{{JID}}">
        </leave>
    </iq>
    '''

    def __init__(self, groupList):
        super(LeaveGroupsIqProtocolEntity, self).__init__(to = YowConstants.WHATSAPP_GROUP_SERVER, _type = "set")
        self.setProps(groupList if type(groupList) is list else [groupList])

    def setProps(self, groupList):
        assert type(groupList) is list and len(groupList), "Must specify a list of group jids to leave"
        self.groupList = groupList

    def toProtocolTreeNode(self):
        node = super(LeaveGroupsIqProtocolEntity, self).toProtocolTreeNode()
        groupNodes = [
            ProtocolTreeNode("group", {
                "id":       groupid
            })
            for groupid in self.groupList
        ]
        node.addChild(ProtocolTreeNode("leave", {"action": "delete"}, groupNodes))
        return node

    @staticmethod
    def fromProtocolTreeNode(node):
        assert node.getChild("leave") is not None, "Not a group leave iq node %s" % node
        assert node.getChild("leave").getAttributeValue("action") == "delete", "Not a group leave action %s" % node
        entity = super(LeaveGroupsIqProtocolEntity, LeaveGroupsIqProtocolEntity).fromProtocolTreeNode(node)
        entity.__class__ = LeaveGroupsIqProtocolEntity
        entity.setProps([group.getAttributeValue("id") for group in node.getChild("leave").getAllChildren()] )
        return entity



from .iq_unregister import UnregisterIqProtocolEntity
from .iq_status_set import SetStatusIqProtocolEntity
from .iq_picture_get import GetPictureIqProtocolEntity
from .iq_picture_get_result import ResultGetPictureIqProtocolEntity
from .iq_pictures_list import ListPicturesIqProtocolEntity
from .iq_picture_set import SetPictureIqProtocolEntity
from .iq_privacy_set import SetPrivacyIqProtocolEntity
from .iq_privacy_get import GetPrivacyIqProtocolEntity
from .iq_privacy_result import ResultPrivacyIqProtocolEntity



from yowsup.structs import ProtocolEntity, ProtocolTreeNode
from yowsup.layers.protocol_receipts.protocolentities  import OutgoingReceiptProtocolEntity
from copy import deepcopy

class MessageProtocolEntity(ProtocolEntity):

    MESSAGE_TYPE_TEXT = "text"
    MESSAGE_TYPE_MEDIA = "media"

    def __init__(self, _type, _id = None,  _from = None, to = None, notify = None, timestamp = None, 
        participant = None, offline = None, retry = None):

        assert (to or _from), "Must specify either to or _from jids to create the message"
        assert not(to and _from), "Can't set both attributes to message at same time (to, _from)"


        super(MessageProtocolEntity, self).__init__("message")
        self._type          = _type
        self._id            = self._generateId() if _id is None else _id
        self._from          =_from
        self.to             = to
        self.timestamp      = int(timestamp) if timestamp else self._getCurrentTimestamp()
        self.notify         = notify
        self.offline        = offline == "1" if offline is not None else offline
        self.retry          = int(retry) if retry else None
        self.participant    = participant

    def getType(self):
        return self._type

    def getId(self):
        return self._id

    def getTimestamp(self):
        return self.timestamp

    def getFrom(self, full = True):
        return self._from if full else self._from.split('@')[0]

    def isBroadcast(self):
        return False

    def getTo(self, full = True):
        return self.to if full else self.to.split('@')[0]

    def getParticipant(self, full = True):
        return self.participant if full else self.participant.split('@')[0]

    def getNotify(self):
        return self.notify
    
    def toProtocolTreeNode(self):
        attribs = {
            "type"      : self._type,
            "id"        : self._id,
        }

        if self.isOutgoing():
            attribs["to"] = self.to
        else:
            attribs["from"] = self._from

            attribs["t"] = str(self.timestamp)

            if self.offline is not None:
               attribs["offline"] = "1" if self.offline else "0"
            if self.notify:
                attribs["notify"] = self.notify
            if self.retry:
                attribs["retry"] = str(self.retry)
            if self.participant:
                attribs["participant"] = self.participant


        xNode = None
        #if self.isOutgoing():
        #    serverNode = ProtocolTreeNode("server", {})
        #    xNode = ProtocolTreeNode("x", {"xmlns": "jabber:x:event"}, [serverNode])


        return self._createProtocolTreeNode(attribs, children = [xNode] if xNode else None, data = None)

    def isOutgoing(self):
        return self._from is None

    def isGroupMessage(self):
        if self.isOutgoing():
            return "-" in self.to
        return self.participant != None

    def __str__(self):
        out  = "Message:\n"
        out += "ID: %s\n" % self._id
        out += "To: %s\n" % self.to  if self.isOutgoing() else "From: %s\n" % self._from 
        out += "Type:  %s\n" % self._type
        out += "Timestamp: %s\n" % self.timestamp
        if self.participant:
            out += "Participant: %s\n" % self.participant
        return out

    def ack(self, read=False):
        return OutgoingReceiptProtocolEntity(self.getId(), self.getFrom(), read, participant=self.getParticipant())

    def forward(self, to, _id = None):
        OutgoingMessage = deepcopy(self)
        OutgoingMessage.to = to
        OutgoingMessage._from = None
        OutgoingMessage._id = self._generateId() if _id is None else _id
        return OutgoingMessage

    @staticmethod
    def fromProtocolTreeNode(node):

        return MessageProtocolEntity(
            node.getAttributeValue("type"), 
            node.getAttributeValue("id"),
            node.getAttributeValue("from"),
            node.getAttributeValue("to"),
            node.getAttributeValue("notify"),
            node.getAttributeValue("t"),
            node.getAttributeValue("participant"),
            node.getAttributeValue("offline"),
            node.getAttributeValue("retry")
            )



from yowsup.layers import  YowProtocolLayer
from .protocolentities import *
from yowsup.layers.protocol_iq.protocolentities import ErrorIqProtocolEntity, ResultIqProtocolEntity
class YowProfilesProtocolLayer(YowProtocolLayer):
    def __init__(self):
        handleMap = {
            "iq": (self.recvIq, self.sendIq)
        }
        super(YowProfilesProtocolLayer, self).__init__(handleMap)

    def __str__(self):
        return "Profiles Layer"

    def sendIq(self, entity):
        if entity.getXmlns() == "w:profile:picture":
            if entity.getType() == "get":
                self._sendIq(entity, self.onGetPictureResult, self.onGetPictureError)
            elif entity.getType() == "set":
                self._sendIq(entity, self.onSetPictureResult, self.onSetPictureError)
            elif entity.getType() == "delete":
                self._sendIq(entity, self.onDeletePictureResult, self.onDeletePictureError)
        elif entity.getXmlns() == "status":
            self._sendIq(entity, self.onSetStatusResult, self.onSetStatusError)
        elif entity.getXmlns() == "privacy":
            self._sendIq(entity, self.onPrivacyResult, self.onPrivacyError)

    def recvIq(self, node):
        pass

    def onPrivacyResult(self, resultNode, originIqRequestEntity):
        self.toUpper(ResultPrivacyIqProtocolEntity.fromProtocolTreeNode(resultNode))

    def onPrivacyError(self, errorNode, originalIqRequestEntity):
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(errorNode))

    def onSetStatusResult(self, resultNode, originIqRequestEntity):
        self.toUpper(ResultIqProtocolEntity.fromProtocolTreeNode(resultNode))

    def onSetStatusError(self, errorNode, originalIqRequestEntity):
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(errorNode))

    def onGetPictureResult(self, resultNode, originalIqRequestEntity):
        self.toUpper(ResultGetPictureIqProtocolEntity.fromProtocolTreeNode(resultNode))

    def onGetPictureError(self, errorNode, originalIqRequestEntity):
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(errorNode))

    def onSetPictureResult(self, resultNode, originalIqRequestEntity):
        self.toUpper(ResultGetPictureIqProtocolEntity.fromProtocolTreeNode(resultNode))

    def onSetPictureError(self, errorNode, originalIqRequestEntity):
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(errorNode))

    def onDeletePictureResult(self, resultNode, originalIqRequestEntity):
        self.toUpper(ResultIqProtocolEntity.fromProtocolTreeNode(resultNode))

    def onDeletePictureError(self, errorNode, originalIqRequestEntity):
        self.toUpper(ErrorIqProtocolEntity.fromProtocolTreeNode(errorNode))



from yowsup.layers.protocol_iq.protocolentities.test_iq import IqProtocolEntityTest
from yowsup.layers.axolotl.protocolentities import SetKeysIqProtocolEntity
from yowsup.structs import ProtocolTreeNode
class SetKeysIqProtocolEntityTest(IqProtocolEntityTest):
    def setUp(self):
        super(SetKeysIqProtocolEntityTest, self).setUp()
        # self.ProtocolEntity = SetKeysIqProtocolEntity
        #
        # regNode = ProtocolTreeNode("registration", data = "abcd")
        # idNode = ProtocolTreeNode("identity", data = "efgh")
        # typeNode = ProtocolTreeNode("type", data = "ijkl")
        # listNode = ProtocolTreeNode("list")
        # for i in range(0, 2):
        #     keyNode = ProtocolTreeNode("key", children=[
        #         ProtocolTreeNode("id", data = "id_%s" % i),
        #         ProtocolTreeNode("value", data = "val_%s" % i)
        #     ])
        #     listNode.addChild(keyNode)
        #
        # self.node.addChildren([regNode, idNode, typeNode, listNode])
